#lang racket/base

(module general racket/base (require (submod "snack/contract.rkt" core))

  (provide (contract-out [ors  ((list*/c string? (listof string?)) . → . string?)]
                         [ands ((list*/c string? (listof string?)) . → . string?)]
                         [indefinite (string? string? ... . → . string?)]
                         [cardinal ([natural/c] [#:zero string?] . →∗ . string?)]
                         [ordinal  (exact-positive-integer? . → . string?)]
                         [n-things (natural/c string? . → . string?)]
                         [plural-suffix (natural/c . → . (or/c "" "s"))]
                         [arity ([natural/c (or/c natural/c +inf.0) string? natural/c]
                                 [#:pluralize? boolean?]
                                 . →∗ . (list/c string? string?))]))
  
  (require "snack/string.rkt"
           "snack/match.rkt"
           (submod "snack/syntax.rkt" core)
           racket/local)
  
  (define (plural-suffix n) (if (n . = . 1) "" "s"))
  (define (n-things n thing) (~a␣ (cardinal n #:zero "no") (~a thing (plural-suffix n))))
  
  (define indefinite-exceptions '("unary" "unique"))

  (define (indefinite noun-clause₀ . rest)
    (define (a-or-an what-follows)
      (if (or (equal? "" what-follows)
              (not (memv (string-ref what-follows 0) '(#\a #\e #\i #\o #\u)))
              (ormap (λ (•) (string-prefix? what-follows •)) indefinite-exceptions))
          "a"
          "an"))
    (apply ~a␣ (a-or-an noun-clause₀) noun-clause₀ rest))

  (define (cardinal n #:zero [zero "none"])
    (if (n . <= . 9)
        (list-ref (list* zero '("one" "two" "three" "four" "five" "six" "seven" "eight" "nine")) n)
        (~a n)))

  (define (ordinal n⁺)
    (if (n⁺ . <= . 10)
        (list-ref 
         '("first" "second" "third" "fourth" "fifth" "sixth" "seventh" "eighth" "ninth" "tenth")
         (sub1 n⁺))
        (~a n⁺ (case (modulo n⁺ 10) [(0 4 5 6 7 8 9) "th"] [(1) "st"] [(2) "nd"] [(3) "rd"]))))
  
  (require (only-in racket/list range))
  (define arity
    (local [(define (too-few  actual) (if (= actual 0)
                                          "none"
                                          (~a␣ "only" (cardinal actual))))
            (define (too-many actual) (cardinal actual))
            (define (interval m M) (ors (map cardinal (range m (add1 M)))))
            (define (at-most m M) (values (~a␣ (cond [(= M 0) "no"]
                                                     [(= m 0) (~a␣ "at most" (cardinal M))]
                                                     [else (~a␣ "only" (interval m M))]))
                                          M))
            (define (at-least m M) (if (< M +inf.0)
                                       (values (interval m M) M)
                                       (values (~a␣ "at least" (cardinal m)) m)))]
      (λ (m M kind actual #:pluralize? [pluralize? #true])
        (define-values (expect up-to) (if (< actual m) (at-least m M) (at-most m M)))
        (list (~a␣ expect (~a kind (if pluralize? (plural-suffix up-to) "")))
              (if (< actual m) (too-few actual) (too-many actual))))))

  (provide (contract-out [code (any/c . → . string?)]))
  (define (code name) (local-require racket/pretty)
    (parameterize ([print-boolean-long-form #true]
                   [pretty-print-show-inexactness #true]
                   [pretty-print-exact-as-decimal #true]
                   [print-as-expression #true])
      (pretty-format name #:mode 'write)))

  (provide (all-from-out (submod "snack/contract.rkt" core) racket/local "snack/string.rkt")
           strings
           (contract-out [comma␣       (string? string? . → . string?)]
                         [parenthesize (string? ...     . → . string?)]
                         [double-quote ((list*/c string? (listof string?)) . → . (listof string?))]))

  (define-syntax-parser strings
    [(_ (~optional (~and prefixed #:prefixed)) [id:id v:expr] ...+)
     #`(begin (define id v) ...
              (provide (prefix-out #,(if (attribute prefixed) #'message: #'||)
                                   (contract-out [id string?] ...))))])
  
  (strings #:prefixed [ellipsis "..."] [period "."])
  
  (define (double-quote   s) (map (λ (•) (~a "\"" • "\"")) s))
  (define (parenthesize . s) (~a "(" (apply ~a␣ s) ")"))
  (define (comma␣ s₀ s₁) (~a␣ (~a s₀ ",") s₁))
  (define-values (ors ands)
    (local [(define-values (“or” “and”) (values "or" "and"))
            (define (combiner with) (λ-match
                                     [(list e₀) e₀]
                                     [(list e₀     e₁) (~a␣ e₀ with e₁)]
                                     [(list e₀ ... e₁) (comma␣ (apply ~a e₀ #:separator ", ")
                                                               (~a␣ with e₁))]))]
      (values (combiner “or”) (combiner “and”)))))


(module ui racket/base (require (only-in (submod ".." general) strings ~a␣))
  (strings [language-name "CSC 104"]
           [language-settings-label (~a␣ language-name " Settings")])
  (strings #:prefixed
           [help "Help"]
           [interactions "Interact"] [history:up "Back"] [history:down "Forth"]
           [definitions "Define"] [layout "Layout"] [record "Record"]
           [insert-image "Image"] [shrink "Shrink"]
           [line-comment "Comment"] [explode "Stack"] [format "Format"] [hug "Hug"])
  (define step: "step:")
  (strings #:prefixed
           [display:underline:next  (~a␣ step: "underline sub-expressions that are about to change")]
           [display:spacing:compact (~a␣ step: "display compactly")]
           [interaction:no-wait     (~a␣ step: "don't wait for response")]))

(module print racket/base (require (submod ".." general))
  (provide message:ellipsis message:period)
  (strings #:prefixed
           [interactions-area "Interactions area."]
           [custom "Custom: "] [prefix-types "Prefix result values with their types"]))

(module step-display racket/base (require (submod ".." general))
  (provide (contract-out [elided ([] [(♯ string?)] . →∗ . string?)]))
  (strings [header (~a␣ "●" "Steps" "●")] [to-step-lead "○"] [step-lead "•"]
           [too-many-steps "too many steps to show"])
  (define (elided [reason #false]) (~a␣ message:ellipsis (or reason ""))))

(module settings-panel racket/base (require (only-in (submod ".." general) strings))
  (strings [output-style "Output Style"] [prefix-type "Prefix result values with their types"]))

(module types racket/base (require (submod ".." general))

  (provide (contract-out [horizontal-positions (listof string?)]
                         [  vertical-positions (listof string?)]
                         [shadings             (listof string?)]))
  
  (define positive "positive")
  (define (non- s) (~a "non-" s))
  
  (strings [image "image"] [list "list"] [boolean "boolean"]
           [text "text"] [character "character"]
           [function "function"] (unary (~a␣ "unary" function)) (binary (~a␣ "binary" function))
           [number "number"]
           [non-negative-number     (~a␣ (non- "negative") number)]
           [non-zero-number         (~a␣ (non- "zero")     number)]
           [positive-number         (~a␣ positive          number)]
           (integer "integer") (natural-number (~a␣ "natural" number))
           [positive-natural-number (~a␣ positive natural-number)])
  
  (define shadings '("solid" "outline"))
  (define-values (horizontal-positions vertical-positions) 
    (values '("left" "right" "center") '("top" "bottom" "center" "baseline")))
  (define (one-of-texts texts) (parenthesize "one of" (ors (double-quote texts))))
  (define position "position")
  
  (strings [smaller-number (~a␣ positive-natural-number "less than 4294967088")]
           [unicode (~a␣ natural-number "at most 55295 or between 57344 and 1114111 inclusive")]
           [color "colour"] [list-of-colors (~a␣ list "of" (~a color "s"))]
           [angle "angle in degrees"]
           [shading (~a␣ "shading mode" (parenthesize (ors (double-quote shadings))))]
           [font-size (~a␣ "font size" (parenthesize "a" number "from 1 to 255 inclusive"))]
           (width "width") (height "height")
           (horizontal-position (~a␣ "horizontal" position (one-of-texts horizontal-positions)))
           (vertical-position   (~a␣   "vertical" position (one-of-texts   vertical-positions)))))

(module syntax racket/base (require (submod ".." general)
                                    "snack/conditional.rkt" "snack/function.rkt")

  #| General CS Terminology |#
  (define-values (parenthesis parentheses open-parenthesis
                              parenthesized grouped clause)
    (values "parenthesis" "parentheses" "open parenthesis"
            "parenthesized" "grouped" "clause"))
  ;
  (define-values (scope:local scope:top-level)
    (values "local" "top level"))
  ;
  (define-values (a-keyword
                  expression literal call variable
                  argument body condition result
                  definition defined undefined parameter)
    (values
     (indefinite "keyword")
     "expression" "literal" "call" "variable"
     "argument" "body" "condition" "result"
     "definition" "defined" "undefined" "parameter"))
  (strings [function-call (~a␣ type:function call)])
  (require (prefix-in type: (only-in (submod ".." types) number text boolean image function)))
  

  ; Wrong things.
  (strings [none "none"])
  (define nothing "nothing")
  (define unknown "something else")
  (define (extra-parts n) (n-things n "extra part"))
  (define (wrong part)
    (define known
      (type-case (syntax-e part)
        [number? type:number] [string? type:text] [boolean? type:boolean] [struct? type:image]
        #;[list? "part"] [else #false]))
    (if known (indefinite known) unknown))
  (strings [incorrect-form "incorrect form"])

  ; Message function contracts.
  (define message₀/c (→ string?))
  (define (message/c argument/c) (argument/c . → . string?))
  (define (message₂/c argument₀/c argument₁/c) (argument₀/c argument₁/c . → . string?))
  (define arity/c (message/c natural/c))
  
  (provide (contract-out [non-top (message/c identifier?)]
                         [bare    (message/c identifier?)]
                         [open-parenthesis-before (message/c symbol?)]
                         [extra-parts (message/c exact-positive-integer?)]
                         [expect-found (message₂/c string? (or/c syntax? string?))]))

  ; Presenting identifiers.
  (define (separate s) (~a " " s " "))
  (define (‘’ id) (separate (~id id)))
  (define (~id id) (~s (syntax-e id)))

  ; Arity.
  (define (arity:expect-found m M kind actual #:pluralize? [pluralize? #true])
    (apply expect-found (arity m M kind actual #:pluralize? pluralize?)))
  (define (extra for total) (expect-found for (extra-parts (sub1 total)))) 
  (define ((prefixer prefix) . s) (apply ~a␣ prefix s))
  (require "snack/match.rkt")
  (match-define (list at-least-one only-one exactly-one just only)
    (map prefixer '("at least one" "only one" "exactly one" "just" "only")))

  ; English.
  (define (plural s) (~a s "s"))
  (define (hyphenate s.0 s.1) (~a s.0 "-" s.1))
  (define (the . ss) (apply ~a␣ "the" ss))
  (define (in-this this) (~a␣ "in this" this))
  (define-values (before after) (values "before" "after"))
  (define-values (the-front-of for) (values (the "front of") "for"))
  (define for-the "for the")
  
  (define (non-top id) (out-of-place (syntax->datum id) (~a␣ "at" (the scope:top-level))))
  (define (not-allowed-here . because-not)
    (apply ~a␣ "not allowed here, because this is not" because-not))
  (define (out-of-place form where-not) (~a␣ "found a use of" form "that is not" where-not))
  
  (define (open-parenthesis-before thing) (indefinite open-parenthesis before (symbol->string thing)))
  (define (bare id) (expect-found (open-parenthesis-before (syntax->datum id)) none))

  (define (found . ss) (apply ~a␣ "found" ss))
  (define (expected . ss) (apply ~a␣ "expected" ss))
  (define (expect-found expect found:syntax/description)
    (comma␣ (expected expect) (~a␣ "but" (found (type-case found:syntax/description
                                                  [syntax? (wrong found:syntax/description)]
                                                  [void? nothing]
                                                  [else found:syntax/description])))))
  (define (missing . ss) (expect-found (apply ~a␣ ss) nothing))
  (define (missings . ss) (arity:expect-found 1 +inf.0 (apply ~a␣ ss) 0))  
  (define ((unlike . ss) part) (expect-found (apply ~a␣ ss) part))
  
  (define definitions (plural definition))
  (define-values (sequence-of-definitions an-expression)
    (values (~a␣ "sequence of" definitions) (indefinite expression)))
  (define for-the-body (~a␣ for-the body))
  (define-values (grouped-definition expression-for-body)
    (values (~a␣ grouped definition) (~a␣ expression for-the-body)))

  
  (define-values (was-defined in-language no-redefine) #;"or a required library"
    (values (~a␣ "was" defined "previously") (~a␣ "is already" defined "in our language")
            (~a "and cannot be re-" defined)))
  (define ((redefine origin) id) (~a␣ (‘’ id) origin no-redefine))

  ; Simple form arity
  ; —————————————————
  
  (provide (contract-out [conditions:less-than-two arity/c]
                         [expressions:not-one      arity/c]
                         [expressions:not-two      arity/c]))
  
  (define-values (conditions:less-than-two expressions:not-one expressions:not-two)
    (values (curry arity:expect-found 2 +inf.0 condition)
            (curry arity:expect-found 1      1 expression)
            (curry arity:expect-found 2      2 expression)))

  ; define
  ; ——————
  (define-values (form:define definition-context variable-name function-name parameter-name)
    (values "define"
            (~a␣ "at" (the scope:top-level) "nor in a sequence of" scope:local definitions)
            (~a␣ variable "name") (~a␣ type:function "name") (~a␣ parameter "name")))
  (define definition-binding
    (~a␣ (~a (indefinite variable-name) ",")
         "or" parenthesized function-name "with its" (plural parameter)))
  (define (expression-after-variable name) (~a␣ expression after (the variable-name (‘’ name))))
  (define expression-for-function-body (~a␣ expression for-the-body "of" (the type:function)))
  (define (name-for . thing) (apply ~a␣ "name for" thing))
  (define the-name-for-the-function (the (name-for (the type:function))))
  (define the-name-for-a-parameter (the (name-for (indefinite parameter))))
  (define name-for-the-parameter (name-for (the parameter)))

  (strings [define:context (out-of-place form:define definition-context)]
           [define:keyword (expect-found definition-binding a-keyword)]
           [define:function:keyword (expect-found the-name-for-the-function a-keyword)])
  
  (provide (contract-out [define:redefine:user     (message/c identifier?)]
                         [define:redefine:non-user (message/c identifier?)]
                         [define:non-binding      (message/c syntax?)]
                         [define:function:non-id  (message/c syntax?)]))
  
  ; Arity.
  (provide (contract-out [define:variable:init:arity (message₂/c natural/c identifier?)]))
  (strings [define:nothing (missing definition-binding)])
  ; • variable
  (define (define:variable:init:arity amount variable)
    (if (= amount 0)
        (missing (indefinite (expression-after-variable variable)))
        (extra (exactly-one (expression-after-variable variable)) amount)))
  ; • function
  (define (function:body:arity amount)
    (if (= amount 0)
        (missing (indefinite expression-for-function-body))
        (extra (only-one expression-for-function-body) amount)))  
  
  (define-values (define:redefine:user define:redefine:non-user)
    (values (redefine was-defined) (redefine in-language)))
  (define define:non-binding (unlike definition-binding))
  (define define:function:non-id (unlike the-name-for-the-function))

  (define (different-parameter than) (indefinite name-for-the-parameter "different than" than))
  ; Functions, including anonymous.
  (provide (contract-out
            [function:body:arity  arity/c]
            [parameters:none (message₂/c identifier? boolean?)]
            [parameters:non-id          (message/c syntax?)]
            [parameters:duplicate       (message/c identifier?)]
            [parameters:local:duplicate (message/c identifier?)]))
  (strings [parameters:keyword (expect-found the-name-for-a-parameter a-keyword)]
           [define:parameters:function (expect-found (different-parameter (the function-name))
                                                     (the function-name "again"))])
  (define parameters:non-id (unlike the-name-for-a-parameter))
  (define-values (parameters:duplicate parameters:local:duplicate)
    (local [(define ((duplicate for) id) (expect-found for (the "duplicate name" (‘’ id))))]
      (values (duplicate (indefinite "unique" name-for-the-parameter))
              (duplicate (different-parameter (the (plural parameter)
                                                   "of" (the "enclosing" type:function)))))))
  (define (parameters:none function anonymous?)
    (missings parameter-name after (if anonymous? (~id function) (the function-name (‘’ function)))))

  ; with
  ; ————
  (strings [with:definitions:none (missing (at-least-one grouped-definition))])
  (strings [with:parts:none       (missing (plural grouped-definition))])
  (provide (contract-out [with:body:arity     arity/c]
                         [with:non-sequence   (message/c syntax?)]
                         [with:non-definition (message/c syntax?)]
                         [with:duplicate      (message/c identifier?)]))  
  (define (with:body:arity amount) (if (= amount 0)
                                       (missing (indefinite expression-for-body))
                                       (extra (only-one expression-for-body) amount)))
  (define with:non-sequence (unlike (plural grouped-definition)))  
  (define with:non-definition (unlike (indefinite sequence-of-definitions)))
  (define with:duplicate (redefine (~a␣ was-defined (in-this sequence-of-definitions))))
    
  ; if else
  ; ———————
  (define-values (form:if form:else final) (values "if" "else" "final"))
  
  (define if-clause               (~a␣ form:if clause))
  (define condition-result-clause (~a␣ (hyphenate condition result) clause))
  (define a-condition-and-result  (ands (list (indefinite condition) result)))
  (define else-result-clause      (~a␣ (hyphenate form:else result) clause))
  (define the-final-clause        (the final clause))
  (define the-condition           (the condition))

  (strings
   [else:context (not-allowed-here the-condition "of" (indefinite if-clause))]
   [if:nothing (missing (indefinite condition-result-clause))]
   [if:else:only (missings "other" clause before (the else-result-clause))]
   [if:clause:only (expect-found (indefinite else-result-clause) none)]
   [if:else:early (found (indefinite else-result-clause "that is not" the-final-clause))]
   [if:else:missing (expected the-final-clause "to be" (indefinite else-result-clause))])
  (provide (contract-out [if:clause:arity arity/c]
                         [if:non-clause (message/c syntax?)]))  
  (define (if:clause:arity amount)
    (case amount
      [(0)  (missing a-condition-and-result)]
      [(1)  (expect-found (indefinite result expression after the-condition) none)]
      [else (extra (just a-condition-and-result) (sub1 amount))]))
  (define if:non-clause (unlike a-condition-and-result))

  ; fun
  ; ———  
  (provide (contract-out [fun:header:bare (message/c identifier?)]))
  (define (fun:header:bare form)
    (arity:expect-found 2 2 (~a␣ parentheses before (~id form)) #:pluralize? #false 1))

  ; id
  ; ——  
  (strings [top:unbound (~a␣ "this" variable "is" undefined)])
  
  ; ⌢:call
  ; ——————  
  (provide (contract-out [call:non:id/fun (message/c (or/c syntax? void?))]
                         [call:nullary    (message/c (or/c identifier? syntax?))]))    
  (define call:non:id/fun (unlike (indefinite type:function after (the open-parenthesis))))
  (define (call:nullary id/fun)
    (missings argument expression after (separate (code (syntax->datum id/fun)))))

  ; big-bang
  ; ————————
  (define initial-model "initial model")  
  (provide (contract-out [big-bang:arity arity/c]))  
  (define (big-bang:arity amount) (missing an-expression for-the initial-model))
  
  ; step
  ; ————
  (define-values (form:step form:hide form:show) (values (ors (list "step" "steps")) "hide" "show"))
  
  (define only-literal-function-arguments (only literal type:function (plural argument)))
  (define-values (hide-clause show-clause) (values (indefinite form:hide clause)
                                                   (indefinite form:show clause)))
  
  (provide
   (prefix-out message: (contract-out [hide:out-of-context message₀/c]
                                      [show:out-of-context message₀/c]
                                      [hide:non:function/call (message/c (or/c syntax? void?))]
                                      [hide:call:non-id (message/c syntax?)]
                                      [hide:call:nullary (message/c identifier?)]
                                      [hide:call:argument:non-literal (message/c syntax?)])))  
  (define-values (hide:out-of-context show:out-of-context)
    (local [(define ((out-of-context context)) (not-allowed-here the-front-of context for form:step))]
      (values (out-of-context hide-clause) (out-of-context show-clause))))  
  (define hide:non:function/call (unlike (indefinite (ors (list type:function function-call)))))
  (define hide:call:non-id call:non:id/fun)
  (define (hide:call:nullary f-id) (missings argument after (~id f-id)))
  (define hide:call:argument:non-literal (unlike only-literal-function-arguments)))


(module error racket/base (require (submod ".." general) (prefix-in type: (submod ".." types)))
  
  (provide (all-from-out (submod ".." general)))
  (strings [argument "argument"] [arguments (~a argument "s")])
  (provide (contract-out [context-description ((or/c symbol? string?) string? . → . string?)]
                         [expect-received                    (string? string? . → . string?)]
                         [for-the-nth-argument (exact-positive-integer? . → . string?)]
                         [number-of-arguments                (natural/c . → . string?)]))

  (define (context-description context description) (~a␣ context ":" description))
  (define (expect-received expect received) (comma␣ (~a␣ "expects" expect)
                                                    (~a␣ "but received" received)))
  (define (for-the-nth-argument position) (~a␣ "for the" (ordinal position) argument))
  (define (number-of-arguments arity) (n-things arity argument)))

(module rewrite racket/base (require (submod ".." error) "snack/lambda.rkt")
  (provide arity
           λ₁
           for-the-nth-argument
           (contract-out [read-syntax:text:image        (any/c         . → . string?)]
                         [read-syntax:illegal-character (any/c string? . → . string?)]))
  (define (read-syntax:text:image        _)
    (context-description "literal text" "cannot contain an image"))
  (define (read-syntax:illegal-character _ c)
    (context-description c "this character is not usable")))

(module stepper racket/base (require (submod ".." general))
  (provide (prefix-out message: (contract-out [local:function-escaping (identifier? . → . string?)])))
  (require (prefix-in type: (submod ".." types)) (only-in (submod ".." error) context-description))
  (define (local:function-escaping f)
    (define name:hofs (list "map" "combine" #;"apply" "repeats" "select" #;"sift" "sort"))
    (context-description
     "stepping" (comma␣ (~a␣ type:function (~symbol (syntax-e f))
                             "was defined during stepping and can only be stepped by calling it")
                        (~a␣ "or as the first argument to" (ors name:hofs))))))

(module problems racket/base
  
  (require (prefix-in type: (only-in (submod ".." types)
                                     function number text list boolean
                                     list-of-colors width height))
           (prefix-in error: (only-in (submod ".." error) expect-received argument))
           (submod ".." general))
  
  (provide code
           (contract-out [predicate-result (symbol? string? string? . → . string?)]
                         [≤-something-of-the (string? string? . → . string?)]
                         [list-length-at-least (exact-positive-integer? . → . string?)]))

  (define (predicate-result for from result)
    (error:expect-received (~a␣ (indefinite type:boolean) "from" from
                                (parenthesize "the" type:function "given to" (~symbol for)))
                           result))

  (define (length-of thing) (~a␣ "length of the" thing))
  (define (list-length-at-least n)
    (if (= n 1) "non-empty list" (~a␣ "list with at least" (n-things n "element"))))
  (strings [length-of-color-list
            (~a␣ "expected" "the" (length-of type:list-of-colors) "to equal the product of"
                 "the" (ands (list (ordinal 2) (ordinal 3))) (~a error:argument (plural-suffix 3))
                 (parenthesize "the" (ands (list type:width type:height))))])
  
  (define-values (“<” “=” “>”) (values "less than" "equal to" "greater than"))
  (define-values (“≤” “≥”) (values (ors (list “<” “=”)) (ors (list “>” “=”))))  
  
  (define (≤-something-of-the something the) (~a␣ type:number “≤” "the" something "of the" the))
  (strings [number-out-of-order (local [(define (≥-the the) (~a␣ type:number “≥” "the" the))]
                                  (≥-the "second argument"))])
  (define (<-length type) (~a␣ type:number “<” "the" (length-of type)))
  (define (≤-length type) (~a␣ type:number “≤” "the" (length-of type)))
  (strings [list-index-too-large (<-length type:list)]
           [list-range-index-too-large (≤-length type:list)]
           [text-index-too-large (<-length type:text)]
           [text-range-index-too-large (≤-length type:text)]))

(module require racket/base (require (only-in (submod ".." general) strings ~a␣))
  
  (provide (prefix-out message: (combine-out a-library-file-string-cannot
                                             a-library-file-string-can
                                             a-new-name-for
                                             variable-not-defined-in-module
                                             not-library-nor-clause
                                             extra))
           message:expect-found message:extra-parts message:none)
  
  (require (prefix-in message: (only-in (submod ".." syntax)
                                        expect-found extra-parts open-parenthesis-before none)))
  
  (define (a-library-file-string-cannot restriction) (~a␣ "a library file name cannot" restriction))
  (define (a-library-file-string-can    restriction) (~a␣ "a library file name can"    restriction))
  
  (strings #:prefixed
           [be-empty "be empty"]
           [start-with-a-slash "start with a slash"]
           [end-with-a-slash "end with a slash"]
           [a-library-file:contain-only "contain only a-z, A-Z, 0-9, -, _, ., space, and slash"]
           [a-library-path "a library path"]
           [after-the-prefix "after the prefix"]
           [a-name-as-a-prefix "a name as a prefix"]
           [require-name "the name of something to require"]
           [just-require-rename "just the name of something to require and a new name"]
           [require-name-or-rename "a name or an old-name/new-name pair"]
           [only-prefix-and-path "only a prefix and a library path"]
           [a-library-path-or-clause "a library path or clause"]
           [a-library-path-or-clause-after-require "a library path or clause after require"]
           [a-single-library-path-or-clause-after-require
            "a single library path or clause after require"]
           [at-least-one-name-after-library-path "at least one name after the library path"])
  
  (define (a-new-name-for id-stx) (~a␣ "a new name for" (syntax->datum id-stx)))
  (define (variable-not-defined-in-module id-stx module-path)
    (~a␣ "the variable or function" (syntax->datum id-stx) "is not defined in" module-path))
  
  (strings #:prefixed

           [a-library-path-after-the-prefix (~a␣ a-library-path after-the-prefix)]
  
           [only-in:bare
            (message:expect-found (message:open-parenthesis-before 'only-in) message:none)]
           [only-in:no-parts
            (message:expect-found a-library-path message:none)]
           [prefix-in:bare
            (message:expect-found (message:open-parenthesis-before 'prefix-in) message:none)]
           [prefix-in:no-parts
            (message:expect-found a-name-as-a-prefix message:none)]

           [no-parts (message:expect-found a-library-path-or-clause-after-require message:none)]
  
           [sub-form-out-of-context "not allowed here, because this is not a require form"]
           [file-string:disallowed-character
            (a-library-file-string-can a-library-file:contain-only)]
           [invalid-module-id "incorrect format for a library path"])
  
  (define (not-library-nor-clause part-stx)
    (message:expect-found a-library-path-or-clause part-stx))
  (define (extra extra-stx)
    (message:expect-found a-single-library-path-or-clause-after-require
                          (message:extra-parts (length (syntax->list extra-stx)))))
  
  (strings #:prefixed
           [file-string:empty (a-library-file-string-cannot be-empty)]
           [file-string:starting-slash (a-library-file-string-cannot start-with-a-slash)]
           [file-string:ending-slash (a-library-file-string-cannot end-with-a-slash)]))
