#lang racket/base

(require racket/require
         (submod "../english.rkt" stepper)
         
         (submod "../snack/contract.rkt" core)
         "../snack/generator.rkt"
         (only-in racket/sequence sequence/c sequence-ref)
         (only-in racket/stream for/stream)

         syntax/parse
         (multi-in
          "../snack"
          ("syntax.rkt"
           "functional.rkt" "boolean.rkt" "conditional.rkt" "iteration.rkt" "match.rkt"))
         (submod "../snack/syntax.rkt" syntax-classes/core)
         (except-in "../snack/list.rkt" find-first)
         
         racket/local)

(require "patterns.rkt" "define.rkt" "context.rkt" "options.rkt"
         (only-in "private/common.rkt" new? new next markup-expression cleanse))
(require (only-in "../language/forms/forms.rkt" compthink:#%app))

(require racket/contract)
(require (only-in racket/list splitf-at))

(provide hiding
         (contract-out [definitions (parameter/c (listof definition/c))] ; Needs review.
                       [program-steps (syntax? #;(listof expression/c)
                                               . → . (listof (sequence/c #:min-count 1 step/c)))]  
                       [error?       predicate/c] ; A step that is/has an error is wrapped in ‘error’.
                       [whole-error? predicate/c] ; Whole expression's reduction produced error?
                       [error-exn        (error? . → . exn:fail?)] ; First sibling error.
                       ; When not ‘whole-error?’, an illustrative expression that can be chunked.
                       [error-expression (error? . → . syntax?)]))

(define-values (hide:if-conditions hide:if-introduction hide:accumulate)
  (values 'if-conditions 'if-introduction 'accumulate))

(define-values (definition/c expression/c) (values any/c any/c)) ; Needs review.

; The step forms ask to chunk these, expecting neither of ‘next’ and ‘new’ self-nested.
(define chunkable/c (and/c (not/c error?) any/c)) ; Needs review.

; For non-whole-error errors, the step forms ask to chunk the error-expression.
(define step/c (or/c error? chunkable/c namespace?))

; ———
(define-macro (parses? pattern:expr ...) (♯syntax-parser [pattern #true] ...))

#| Steps: |#

#;(contracts [(steps any/c [#:top? boolean? #:newed any/c]) →∗ stream?]
             [value-only? (parameter/c boolean?)] [internal? (parameter/c boolean?)]
             [(reducible? unreduced?) → boolean?] [(decompose unreduced?) → decomposition?])
#;(define (reducible? e) (hole? (decomposition-context (decompose e))))

#;(define (transplant expression+ context)
    (define-values (un-newed re-newer) (if (new? expression+)
                                           (values (markup-expression expression+) new)
                                           (values expression+ (λ• •))))
    (re-newer (type-case context
                [list? (map transplant un-newed context)]
                [hole? (next un-newed)]
                [else un-newed])))

#|  • of a top-level sequence |#

(define (program-steps program)
  (define definition-context '(not-a-definition))
  (let loop ([program (syntax->list program)])
    (match program
      [`(,top-level ,program ...)
       (define steps
         (generator (λ (yield)
                      (define definition-context′
                        (join (syntax-parse (values #;cleanse (first definition-context))
                                [(~is Define) (list this-syntax)]
                                [_ '()])
                              (rest definition-context)))
                      (with-locals (reverse definition-context′)
                        (set! definition-context (adjoin (void) definition-context′))
                        ((stepper top-level #:top? #true) yield)))))
       (adjoin (for/stream ([step steps])
                 (set! definition-context (list* (key:next:cleanse step) (rest definition-context)))
                 (key:next→next step))
               (loop program))]
      ['() (list (list (current-namespace)))])))

(define key:next 'compthink:next)
(define (key:next? stx) (syntax-property stx key:next))
(define (key:next:set stx) (syntax-property stx key:next #true))
(define (key:next:unset stx) (syntax-property-remove stx key:next))
(define (key:next:cleanse s)
  (if (error? s) s (let key:next:cleanse ([s s]) (if (key:next? s) 
                                                     (key:next:cleanse (key:next:unset s))
                                                     (syntax-parse s
                                                       [(s ...) (syn (map key:next:cleanse (α s)))]
                                                       [_ s])))))
(define (key:next→next s)
  (if (error? s) s (let key:next→next ([s s]) (if (key:next? s)
                                                  (next (key:next→next (key:next:unset s)))
                                                  (syntax-parse s
                                                    [(s ...) (map key:next→next (α s))]
                                                    [_ (syntax-e s)])))))

; Hiding
(define (∈-hide? e) (member? (if (syntax? e) (syntax->datum e) e) (hide)))
(define (hiding-call? a-call)
  (ormap ∈-hide? (list a-call (♯syntax-parse a-call [(~and (~is-not For-step) :Call) (α name)]))))
(define (for:hide?) (∈-hide? hide:accumulate))
(define (if:hide-conditions?)   (or (if:hide-introduction?) (∈-hide? hide:if-conditions)))
(define (if:hide-introduction?) (∈-hide? hide:if-introduction))

#|  • of an expression |#

(define (split l)
  (define-values (•s ⊚s) (splitf-at l (parses? (~is •))))
  (list* •s (cond [(empty? ⊚s) ⊚s] [else (define split′ (split (rest ⊚s)))
                                         (list* (list (first ⊚s) (first split′))
                                                (rest split′))])))

(define/syntax-parse ⊙ (hole))
(define-macro (⊕ r:expr ... r′:expr) (Hole (λ () r ... r′)))
(define maybe-hole (syntax-parser [(~is ⊚) (hole)] [_ this-syntax]))
(struct Context (builder unplugged) #:transparent)

(define-syntax-parser define-parser #:literals (syntax)
  [(_ id:id pattern:expr (~optional (syntax data:expr)) result:expr ... final:expr)
   #'(define id (syntax-parser [pattern (~? (syntax-parse #'data result ... [_ (⊕ final)])
                                            (~@ result ... final))]
                               [_ #false]))])
(define-syntax-parser build [(~or* _:id (_ . parts))
                             #:with builder (datum->syntax this-syntax 'builder)
                             #'{~? ((α builder) . parts) (α builder)}])
(define-syntax-parser build₁ [(_ {~optional builder′:expr} e:id)
                              #:with builder (datum->syntax this-syntax 'builder)
                              #'(Context {~? builder′ (α builder)} (list (α e)))])
(define (♯builds es [post values])
  (match-define (list •s (list ⊚⋯ •s⋯) ...) (split es))
  (and (not (empty? ⊚⋯)) (Context (λ es (post (append •s (append-map list* es •s⋯)))) ⊚⋯)))

#| Decomposers |#

(define (♯chase an-id) (syntax-parse (♯binding an-id)
                         [:Define-Variable #:when (identifier? (α value)) (♯chase (α value))]
                         [_ this-syntax]))

; id
; ——
; How do we get here?
;   What's considered ⊚ for an id, and which spots handle ids themselves.
;   What's considered literal?
(define-parser id:context r:id
  (⊕ (let loop ([♯b (♯chase (α r))])
       (syntax-parse ♯b
         [(~and (~or* :Define-Function :Define-Variable) (~is (α value) Literal)) (α value)]
         [:Define-Function (Error (message:local:function-escaping (α name)))]
         [#false (Value (α r))]))))

(require (only-in "../language/without-gui.rkt" prepend))
(define-parser for/list:context :For/List-step
  #'s [s:⊚ (build₁ s)]
  (skipper
   (λ (reducer) (cond [(for:empty-sequence? (α s)) (build-List (list))]
                      [else (define-values (s₀ s′) (for:split-sequence (α s)))
                            (slist #'prepend (reducer (substitute (α parameter) (list s₀) (α body)))
                                   (build s′))]))))

(define-parser for:context :For-step
  #'es [(e ...) #:do [(define b (♯builds (α e) (α builder∗)))] #:when b b]
  (skipper
   (λ (reducer)
     (if (for:sequence? (α s))
         (if (α as)
             (cond [(for:empty-sequence? (α s)) (α a⒮)]
                   [else (define-values (s₀ s′) (for:split-sequence (α s)))
                         (define a ((if (for:hide?) reducer values)
                                    (substitute (α parameter) (list* s₀ (α as)) (α body))))
                         (build s′ a)])
             (Syntax-Error (α accumulator-stub)))
         (Syntax-Error (α sequence-stub))))))
;
(define (for:empty-sequence? e-stx) (case (value←Syntax e-stx) [("" ()) #true] [else #false]))
(define (for:sequence? e-stx) ((∨ string? list?) (value←Syntax e-stx)))
(define (for:split-sequence e-stx)
  (define v (value←Syntax e-stx))
  (define-values (s₀ s′) (if (string? v)
                             (values (substring v 0 1) (substring v 1))
                             (values (first v)         (rest v))))
  
  (values (datum->syntax e-stx (literal←value s₀))
          (datum->syntax e-stx (literal←value s′))))

(define-parser define:context :Define-step #'body [body:⊚ (build₁ body)] void)

(define-parser step:context :Step-step #'e (α e))

(define-parser with:context :With-step
  #'((definition ...) body)
  [([•definition:• ... d:⊚ definition ...] _)
   (scope (α •definition) (build₁ (build (α •definition) (α definition)) d))]
  [([•definition:• ...] body:⊚)
   (scope (α •definition) (build₁ body))]
  (α body))

(define-parser block:context :Block-step
  #'es [(e₀:⊚ e′ ...) (build₁ (λ• (build • (α e′))) e₀)]
  (syntax-parse this-syntax
    [((~is •) ... ⊚e:⊚ e′ ...+) (build (α ⊚e) (α e′))]
    [((~is •) ... e′) (α e′)]
    [() (Value (build))]))

(define-parser or/and:context :Or/And-step
  #'non-neutral [c₀:⊚ (build₁ c₀)]
  (syntax-parse this-syntax
    [(~is boolean) this-syntax]
    [_ (Syntax-Error (α non-boolean-stub))]))

(define-parser if:context :If-step
  #'((neutral ...) non-neutral) [(() c:⊚) #:when (not (if:hide-conditions?)) (build₁ c)]
  (peeker
   (λ (step₁) (syntax-parse (α non-neutral)
                [(~or* #true (~is Else)) (α non-neutral-result)]
                [(~is Literal) (Syntax-Error (α non-boolean-stub))]
                [c (match (step₁ (α c))
                     [(error exn c) (error exn (build c))]
                     [c (if (if:hide-conditions?) (step₁ (build c)) (build c))])]))))

(define-parser arithmetic:context :Arithmetic-step (⊕ (Value this-syntax)))
;
(define (name-function-value f-value f-expr)
  (procedure-rename f-value (if (identifier? f-expr) (syntax-e f-expr) '|an anonymous function|)))
;
(define-parser call:context :Call-step
  (or (♯builds (α argument) (λ (suffix) (build-Call (α f) (append (α prefix) suffix))))
      (⊕ (skipper (λ (reducer)
                    (define §g (α g′))
                    (syntax-parse this-syntax
                      [(:h²of-prefix ℓ₁:List ℓ₂:List)
                       #:when (or (α cross) (= (α ℓ₁.the-length) (α ℓ₂.the-length)))
                       ((α ϕF) (α ϕf) §g (α ℓ₁.element) (α ℓ₂.element))]
                      ;
                      [(:Repeats     _:Functional seed n)
                       #:do [(define n′ (value←Syntax (α n)))]
                       #:when (exact-nonnegative-integer? n′) ((α ϕf) §g (α seed) n′)]
                      [(:Map/Combine (~is Functional) ℓ:List) ((α ϕf) §g (α ℓ.element))]
                      [(:Select/Sort g:Functional ℓ:List)
                       (define ϕg (∘ ⊥/value←≈literal reducer §g))
                       (Value ((α §f) (name-function-value (if (α unary)
                                                               (λ (e)     (ϕg (literal←value e)))
                                                               (λ (e₀ e₁) (ϕg (literal←value e₀)
                                                                              (literal←value e₁))))
                                                           #'g)
                                      #'ℓ))]
                      ;
                      [_ ; All above must bind f in case of fall-through due to guards.
                       (define do-substitution (substitution (α argument)))
                       (define make-call (called-with (α argument)))
                       (define ♯chased (or (and (identifier? (α f)) (♯chase (α f))) (α f)))
                       (syntax-parse ♯chased
                         [v:Define-Variable (make-call (α v.value))]
                         [f:Define-Function
                          (do-substitution this-syntax (with-local-function (α f.name) (α f.parameter)
                                                         #false (make-call (α f.name))))]
                         [(~is Fun) (do-substitution this-syntax #'(compthink:#%app f argument ...))]
                         [_ (Value (make-call (α f)))])]))))))

(define ((called-with arguments) f) (build-Call f arguments))
(define ((substitution arguments) f-definition error-representative)
  (define/syntax-parse (~or* :Fun :Define-Function) f-definition)
  (if (same-length? (α parameter) arguments)
      (substitute   (α parameter) arguments (α body))
      (Runtime-Error error-representative)))
                     
#| Stepper |#

(define (mark-next context redexes) #;(transplant e-newed context)
  (plug context (map key:next:set (map redex-expression redexes))))

(define-values (value? internal?) (values (make-parameter #false) (make-parameter #false)))

(define ((stepper top #:top? [top? #false]) yield)

  (define (reduced e) (parameterize ([internal? #true] [value? #true]) (steps e)))
  (define (step₁ r′) (sequence-ref (parameterize ([internal? #true] [value? #false])
                                     (generator (stepper r′))) 1))    
  (define reduce (reducer step₁ reduced))
  (define (blackbox? whole-context r-e reduced-once)
    (and (neither (value?) (error? reduced-once))
         (hiding-call? r-e) (top? . ⇒ . (not (same-call? top r-e)))))
  (define skip-one? (parses? (~true (if:hide-introduction?) (~is If-step))))
  (define (end? e) (♯syntax-parse e [(~or* (~is •) (~true (error? e) _)) #true]))
  (define (same-call? e e′) (♯syntax-parse (list e e′) [((f:id . _) (f′:id . _)) (eq? (α f) (α f′))]))
  (define error-unwrap (λ-identity/match [(error _ expression) expression]))
  (define error-wrapper (λ-match [(error exn _) (curry error exn)] [_ values]))
  
  (define (steps e #:newed [e-newed e])
    (define-macro (yield-e e′) (yield (if (internal?) e e′)))
    (match e
      [(? end?) (if (value?) e (yield-e e #;e-newed))]
      [(app decompose (decomposition context redexes))
       (unless (value?) (yield-e (mark-next context redexes)))
       (define tabulated
         (tabulate new error-unwrap
                   (λ (r+ r) (cond [(blackbox? context (redex-expression r) r+) (reduced r+)]
                                   [(skip-one? r+) (step₁ r+)]
                                   [else r+]))
                   reduce
                   redexes))
       (define error-wrap (error-wrapper (find-first error? (third tabulated))))
       (define in-context (plugger context))
       (steps (error-wrap (in-context (second tabulated)))
              #:newed (error-wrap (in-context (first tabulated))))]))
  (steps top))

#| Reducer |#

(define literal? (parses? (~is Literal)))
(define ≈syntax/c (or/c syntax? error?))
(define ≈literal/c (or/c literal? error?))

; Only for reducers.
(define/contract ReRaise (error? . → . none/c) (∘ raise error-exn))
(define/contract Syntax-Error  (syntax? . → . error?) (∘ Eval syntax->datum))
(define/contract Runtime-Error (any/c #;syntax? . → . error?) Eval  #;(∘ Eval syntax->datum))
; For unique to stepping: local function escaping.
(define/contract Error (string? . → . error?) message→error)

; Only for reducers.
(define/contract (literal←value value) ((not/c ≈syntax/c) . → . literal?)
  (type-case value
    [list? (build-List (map literal←value value))]
    [procedure? (object-name value)]
    [else value]))
;
(define (⊥/value←≈literal ≈literal) (≈literal/c . → . (not/c ≈syntax/c))
  ((if (error? ≈literal) ReRaise value←Syntax) ≈literal))

; For step depending on knowing a user value (repeats number, accumulate sequence).
; Only for reducers.
(define/contract value←Syntax (literal? . → . (not/c ≈syntax/c))
  (∘ Eval syntax->datum))

; non-user-id (including in fn position, possibly hof non-algebraic or hof pass-through by form),
;  and void from (block)
; Only for reducers.
(define/contract (Value stx) (syntax? . → . ≈literal/c)
  (define result (Eval stx #;(syntax->datum stx)))
  (if (error? result) result (datum->syntax stx (literal←value result))))

(define reducer/c (syntax? . → . ≈syntax/c))
(define delayed-reduction/c (→ (or/c ≈syntax/c (reducer/c . → . ≈syntax/c))))

(define-struct/contract redex
  ([definitions (listof syntax?)] [expression syntax?] [reducer delayed-reduction/c]) #:transparent)

(struct reduction (λ) #:transparent #:property prop:procedure 0)
(struct peeker  reduction () #:transparent)
(struct skipper reduction () #:transparent)
(define/contract ((reducer step₁ reduced) a-redex)
  (reducer/c reducer/c . → . (redex? . → . ≈syntax/c))
  (match-define (redex ds r λr) a-redex)
  (with-locals ds
    (define λ/r (λr))
    (cond [(peeker?  λ/r) (λ/r step₁)]
          [(skipper? λ/r) (λ/r reduced)]
          [else λ/r])))

#| Decomposer |#

(define-struct/contract Hole ([reduce-thunk delayed-reduction/c]) #:transparent)
(define-struct/contract scope ([locals (listof syntax?)] [context Context?]) #:transparent)
(define-struct/contract decomposition ([context syntax?] [redexes (listof redex?)]) #:transparent)

(define/contract (decompose expression) (syntax? . → . decomposition?)
  (match-define (and where (or (scope _ context) context)) (look expression))
  (if (Hole? context)
      (decomposition (α ⊙) (list (redex (definitions) expression (Hole-reduce-thunk context))))
      (with-locals (if (scope? where) (scope-locals where) '())
        (match-define (list (decomposition contexts redexes) ...)
          (map decompose (Context-unplugged context)))
        (decomposition (apply (Context-builder context) contexts) (apply join redexes)))))

(define/contract (look e) (syntax? . → . (or/c Context? scope? Hole?))
  (ormap (λ (c) (c e)) (list id:context define:context or/and:context if:context for:context
                             for/list:context step:context with:context block:context
                             arithmetic:context call:context)))
