#lang racket/base

; ✪ snack collection
; ✪ defines #:with
; ✪ submods

(provide Eval (struct-out error) whole-error? message→error)

; Connect to language via namespace reflection
; ————————————————————————————————————————————
; Usage within this module is to determine if an identifier is bound to a function,
;  where the function is not being managed by the stepper's environment (need to
;  describe details of “managed”).
(define (Eval r) (with-handlers ([exn:fail? whole-error]) (eval r)))

; Non-list non-function values
; ————————————————————————————
; Void and boxes are experimental and in-progress.
; Should boxes be treated similarly to compound function expressions?
(require (only-in (submod "private/common.rkt" kinds) image?)
         (only-in "../language/box.rkt" Box?))
(define datum? (∨ boolean? string?
                  real? image?
                  void? Box?))

; Language functions bindings with special stepper handling
; —————————————————————————————————————————————————————————
; Language function aliases refer to distinct functions (so errors can refer to the alias name),
;  thus all must be recognized.
(require (prefix-in ct- (combine-in
                         ; compound literal
                         (only-in "../language/without-gui.rkt" list)
                         ; algebraic rules, and/or marshalling
                         (only-in "../language/without-gui.rkt"
                                  map
                                  combine apply
                                  repeats
                                  parallel cross)
                         ; marshalling
                         (only-in "../language/without-gui.rkt"
                                  only select sift
                                  sort)
                         ; compound literals, potential algebraic rules, marshalling
                         (only-in "../language/without-gui.rkt" compose ∘ >∘ •)
                         ; considering collapsing mutliple arithmetic steps
                         (only-in "../language/without-gui.rkt" + * - / inc dec))))

(provide Literal (rename-out [Reduced •] [Unreduced ⊚]))

(module error racket/base (provide (except-out (all-defined-out) ∅))
  (struct error (exn expression) #:transparent) #;{exn:fail? any/c}
  (define ∅ #'⌢)
  (define (whole-error exn) (error exn ∅))
  (define (whole-error? step) (and (error? step) (equal? (error-expression step) ∅)))
  (define (message→error message) (error (exn:fail message (current-continuation-marks)) ∅)))
(require 'error)

(require (submod "../snack/syntax.rkt" core)
         (submod "../snack/syntax.rkt" syntax-classes/core)
         "../snack/function.rkt")

(require (for-meta -1 (only-in "../language/forms/forms.rkt"
                               define-name-shape
                               fun-shape callable
                               or/and-shape
                               if-shape else-shape
                               with/local-shape
                               Block
                               for-form for/list-form
                               Each-clause With-clause List Variable Single)))
(require (only-in "../language/forms/forms.rkt" with-local-function))

(define-literal-set functional (ct-compose ct-∘ ct->∘ ct-•))
(define-literal-set arithmetic (ct-+ ct-* ct-- ct-/ ct-inc ct-dec))

(define-syntax-class Unreduced (ρ (~is-not Reduced)))
(define-syntax-class Reduced
  (ρ (~or* (~is Literal) (~is Define-Function)))
  (ρ (~and :Define-Variable (~is (α body) Literal))))
(define functional? (literal-set->predicate functional))
(define-syntax-class Literal #:attributes ()
  (ρ (~or* (~true (datum? (syntax-e this-syntax)) _)))
  (ρ (~is Else))
  (ρ (~and (~is id) (~true (procedure? (Eval this-syntax)))))
  (ρ (~is Fun))
  (ρ (~and :List (~parse ((~is Literal) ...) (α elements))))
  (ρ ((~and f:id (~true (functional? (α f)))) (~is Literal) ...)))

(provide Arithmetic-step)
(define-syntax-class Arithmetic-step (ρ (~and (~is Call-step) (~is Arithmetic-literal))))
(define arithmetic? (literal-set->predicate arithmetic))
(define-syntax-class Arithmetic-literal
  (ρ (~is number))
  (ρ (f:id ~rest (~true (arithmetic? (α f)) ((~is Arithmetic-literal) ...)))))

(provide Or/And-step For-step For/List-step Block-step Step-step
         With-step with-local-function If-step Else
         Call-step Functional Map/Combine Fun Repeats Select/Sort
         h²of-prefix h¹of-prefix f-prefix
         Call build-Call List build-List
         with-local-function)

#| Define |#
(provide Define Define-Variable Define-Function Define-step)
(define-syntax-class Define
  #:attributes {binder name named-variable? named-function? [parameter 1] body}
  (ρ :define-name-shape))
(define-syntax-class Define-Function #:attributes {name value body [parameter 1]}
  (ρ :Define #:when (α named-function?) #:with value (α name)))
(define-syntax-class Define-Variable #:attributes {value body}
  (ρ :Define #:when (α named-variable?) #:with value (α body)))
(define-syntax-class Define-step #:attributes {builder body}
  (ρ (~and :Define (define . _)) #:attr builder (λ (body) (slist (α define) (α binder) body))))

#| Fun Call |#
(define-syntax-class Fun  #:attributes {name (parameter 1) body} (ρ :fun-shape))
(define-syntax-class Call #:attributes {name} (ρ (:callable . _)))
(define (build-Call f es) (slist* f es))
(define ((build-Caller f) . es) (build-Call f es))
(define-syntax-class Call-step #:attributes {f [prefix 1] g′ [argument 1]}
  (ρ ((~and (~seq f prefix ...)
            (~or* :h²of-prefix :h¹of-prefix (~is f-prefix)))
      argument ...)
     #:attr g′ (build-Caller (α g))))

#| Or/And |#
(define-syntax-class Or/And-step #:attributes {non-neutral non-boolean-stub builder}
  (ρ :or/and-shape
     #:with ((~same? neutral neu) ... ~rest (~and (non-neutral delayed-condition ...)
                                                  (~or* ((~not (~same? neu)) . _) (_))))
     (α condition)
     #:do [(define prefix (λ (suffix) #`(name neutral ... . #,suffix)))]
     #:attr builder (λ (non-neutral) (prefix (slist* non-neutral (α delayed-condition))))
     #:with §neu (α neu)
     #:attr non-boolean-stub (prefix #`(non-neutral §neu))))

#| If |#
(define-syntax-class If-step
  #:attributes {[neutral 1] non-neutral non-neutral-result builder non-boolean-stub}
  (ρ :if-shape #:with {(~and neutral [#false _]) ...
                       [(~and non-neutral (~not #false)) non-neutral-result]
                       delayed-clause ...}
     (α clauses)
     #:attr non-boolean-stub #'(name [non-neutral #false] [else-name #false])
     #:attr builder (λ (non-neutral) #`(name [#,non-neutral non-neutral-result] delayed-clause ...))))
(define-syntax-class Else #:attributes () (ρ :else-shape))

#| With Local |#
(define-syntax-class With-step #:attributes {[definition 1] body builder}
  (ρ :with/local-shape
     #:do [(define (namer parts) (slist* #'name parts))
           (define Ds (syntax-parser [(ds ...) (if (α with?) (α ds) (list (α ds)))]))]
     #:attr builder (syntax-parser∗
                     [([d ...] [d′′ ...])
                      (syntax-parser [d′ (namer (append (Ds #'[d ... d′ d′′ ...]) (list #'body)))])]
                     [(body)
                      (namer (append (Ds #'[definition ...]) (list #'body)))])))

#| Call, HoFs |#

(define-splicing-syntax-class f-prefix #:attributes {} (ρ (~seq _)))

(define-splicing-syntax-class h¹of-prefix #:attributes {g} (ρ (~seq (~is H¹oF) g:Functional)))
(define-syntax-class H¹oF #:attributes {}
  (ρ (~or* (~is Map/Combine) (~is Select/Sort) (~is Repeats))))

(define-splicing-syntax-class h²of-prefix #:attributes {ϕF cross ϕf g}
  (ρ (~seq :H²oF :Map g:Functional)))
(define-syntax-class H²oF #:attributes (cross ϕF)
  (ρ (~or* (~capture ct-parallel ϕF) (~and (~capture ct-cross ϕF) cross))))

(define-syntax-class Map #:attributes (ϕf)
  (ρ (~literal ct-map) #:attr ϕf (∘ build-List ct-map)))

(define-syntax-class Map/Combine #:attributes (ϕf)
  (ρ (~or* :Map (~capture ct-combine ϕf) (~capture ct-apply ϕf))))

(define-syntax-class Repeats #:attributes (ϕf)
  (ρ (~literal ct-repeats) #:attr ϕf (∘ build-List ct-repeats)))

(define-syntax-class Select/Sort #:attributes (unary §f)
  #:literals (ct-only ct-select ct-sift ct-sort)
  (ρ (~or* (~and unary (~or* f:ct-only f:ct-select f:ct-sift))
           f:ct-sort)
     #:attr §f (build-Caller (α f))))

(define-syntax-class Functional #:attributes {} (ρ (~or* (~is id) (~is Reduced))))

(define build-List (syntax-parser [es #:with list (datum->syntax #'ct-list 'list) #'(list . es)]))


#| Block |#
(define-syntax-class Block-step #:attributes (builder es)
  (ρ :Block #:do [(define (namer parts) (slist* (α name) parts))]
     #:attr builder (case-lambda [() (namer (list))] [(e es) (namer (list* e es))])))

; Step
(define-syntax-class Step-step #:attributes (e) #:datum-literals (step steps)
  (ρ ((~or* name:step name:steps) e:expr)))

#| fors |#
(define-syntax-class For/List-step
  #:attributes ((parameter 1) s body builder)
  (ρ :for/list-form
     #:with (parameter ...) #'parameters
     #:attr builder (λ (new-s) ((α namer) (map (syntax-parser
                                                 [:Each-clause ((α each-builder) new-s)]
                                                 [_ this-syntax])
                                               (α clause))))))
(define-syntax-class For-step
  #:attributes {[parameter 1] es s a⒮ body as builder sequence-stub accumulator-stub builder∗}
  (ρ :for-form
     #:with (~or* (~and _:Variable a-id⒮′:Single)
                  (~and (_:id ...) a-id⒮′:List))
     #'a-id⒮
     #:with (~or* a⒮′:List a⒮′) #'a⒮
     #:attr as (and (or (not (α a-id⒮′.the-length))
                        (equal? (α a-id⒮′.the-length)
                                (α    a⒮′.the-length)))
                    (if (identifier? #'a-id⒮) (list #'a⒮) (α a⒮′.element)))
     #:with (parameter ...) #'parameters
     #:attr builder (λ (new-s new-a⒮ [body′ #'body])
                      ((α namer) (map (syntax-parser [:With-clause ((α with-builder) new-a⒮)]
                                                     [:Each-clause ((α each-builder) new-s)]
                                                     [_ body′])
                                      (α clause))))
     #:attr builder∗
     (λ (es) (apply (α builder) (case (α first-type) [(each) es] [else (reverse es)])))
     #:attr    sequence-stub ((α builder) #'s                          #false #false)
     #:attr accumulator-stub ((α builder) (build-List '(#false #false)) #'a⒮  #false)))
