#lang racket/base (provide compthink:require compthink:only-in compthink:prefix-in)

#| Expansion Time |#

(require (submod "../../snack/syntax.rkt" core)) ; Standard imports.

(begin-for-syntax (require "syntax-error-context.rkt") ; Shared state for syntax errors.
                  ; Generic parser that sets context and checks for ⌢:bare/weird at the end.
                  (require (submod "../../snack/syntax.rkt" core))
                  (define-syntax-parser parser/⊥
                    [(_ part ...+) #'(λ (stx) (parameterize [(context stx)]
                                                (syntax-parse stx part ... [_ (⌢:bare/weird)])))]))

; Will lazily require much of the static error handling.
(begin-for-syntax
  (require racket/lazy-require)
  (lazy-require [(submod "syntax-error.rkt" require) (syntax-error ⌢:bare/weird)]))

(begin-for-syntax ; Standard definitions.
  (require "common.rkt"))

#| ——— |#

(module error racket/base (provide ⌢:require check-module-form parse-only-in parse-prefix-in)

  (require (only-in (submod "syntax-error.rkt" require) syntax-error ⌢:bare/weird)
           "common.rkt"
           (submod "../../english.rkt" require)
           syntax/parse syntax/stx racket/list)
  (require (for-template (only-in racket/base require prefix-in only-in)))

  (define (⌢:require stx)
    (syntax-parse stx
      [_ #:when (not (top?)) (syntax-error this-syntax "step out of context")]
      [(_) (syntax-error "no parts" #;message:no-parts)]
      [(_ something-else) (syntax-error "something else"
                                        #;(message:not-library-nor-clause #'something-else)
                                        #'something-else)]
      [(_ _ . extra) (syntax-error "extra" #;(message:extra #'extra) #'extra)]
      [_ (⌢:bare/weird)]))

  #;(define (⌢:call stx #:nullary-fun? [nullary-fun? #false])
      (define (syntax-error message/r [about (void)])
        (raise-syntax-error (string->symbol message:function-call)
                            (if (procedure? message/r) (message/r about) message/r)
                            stx))
      (syntax-parse stx
        [(_ f) #:when (or (identifier? #'f) nullary-fun?) (syntax-error message:call:nullary #'f)]
        [(_ part . _) (syntax-error message:call:non:id/fun #'part)]
        [(_)          (syntax-error message:call:non:id/fun (void))]))
  
  (define (check-module-form stx module-stx)
    ((if (identifier? module-stx) check-module-id-form check-module-string-form) stx module-stx))
  ;
  (define (check-module-id-form stx module-id-stx)
    (unless (module-path? (syntax-e module-id-stx))
      (syntax-error stx message:invalid-module-id module-id-stx)))
  ;
  (define (check-module-string-form stx module-string-stx)
    (define s (syntax-e module-string-stx))
    (unless (regexp-match #rx#"^[-a-zA-Z0-9_. ]+(/+[-a-zA-Z0-9_. ]+)*$" s)
      (syntax-error stx (cond [(string=? "" s) message:file-string:empty]
                              [(regexp-match #rx"^/" s) message:file-string:starting-slash]
                              [(regexp-match #rx"/$" s) message:file-string:ending-slash]
                              [else message:file-string:disallowed-character])
                    module-string-stx)))

  ; Association list of phase 0 exports of module-path.
  (define (exports module-path)
    ; The second argument should make it so that the module is always declared after this,
    ;  so this is kind of a weird function. Need to do this to be able to use module->exports.
    (module-declared? module-path #t)
    (define-values (vars stxes) (module->exports module-path))
    (define phase-0-vars (assoc 0 vars))
    (define phase-0-stxes (assoc 0 stxes))
    (cond [(and phase-0-vars phase-0-stxes) (append (rest phase-0-vars) (rest phase-0-stxes))]
          [phase-0-vars (rest phase-0-vars)]
          [phase-0-stxes (rest phase-0-stxes)]
          [else (list)]))
  ;
  ; Is syntax object id exported by module-path?
  (define (exported? id module-path) (assoc (syntax->datum id) (exports module-path)))
  ;
  ; Raise an error if id is not exported by module-path at phase 0.
  (define (check-exported id module-path stx)
    (unless (exported? id module-path)
      (syntax-error stx (message:variable-not-defined-in-module id module-path) id)))

  #| Delegated parsing of ‘only-in’ and ‘rename-in’. |#
  ;
  ; Takes an entire require form as input.
  ; The sub-form should already have been checked to match literal compthink:only-in.
  (define (parse-only-in stx)
    (syntax-parse stx
      [(_ bare:id . _)   (syntax-error stx message:only-in:bare     #'bare)]
      [(_ (~and (_) it)) (syntax-error stx message:only-in:no-parts #'it)]
      
      [(_ (_ module:id))
       (syntax-error stx (message:expect-found message:at-least-one-name-after-library-path
                                               message:none))]
      
      ; Correct syntax
      [(_ (_ m-id:id maybe-spec ...+))
       (check-module-id-form stx #'m-id)
       (define m (syntax->datum #'m-id))
       (for ([id/pair (stx->list #'(maybe-spec ...))])
         (syntax-parse id/pair
           [ident:id (check-exported #'ident m stx)]
           [(e-id new-id extra ...)
            (unless (identifier? #'e-id)
              (syntax-error stx (message:expect-found message:require-name #'e-id) #'e-id))
            (check-exported #'e-id m stx)
            (unless (identifier? #'new-id)
              (syntax-error stx (message:expect-found (message:a-new-name-for #'e-id) #'new-id)
                            #'new-id))
            (define number-of-extra-parts (length (stx->list #'(extra ...))))
            (unless (zero? number-of-extra-parts)
              (syntax-error stx (message:expect-found message:just-require-rename
                                                      (message:extra-parts number-of-extra-parts))
                            this-syntax))]
           [_ (syntax-error stx (message:expect-found message:require-name-or-rename id/pair)
                            id/pair)]))
       (syntax/loc stx (require (only-in m-id maybe-spec ...)))]

      [(_ (_ non-id . _)) (syntax-error stx (message:expect-found message:a-library-path #'non-id)
                                        #'non-id)]))
  
  (define (parse-prefix-in stx)
    (syntax-parse stx
      [(_ bare:id . _)   (syntax-error stx message:prefix-in:bare     #'bare)]
      [(_ (~and (_) it)) (syntax-error stx message:prefix-in:no-parts #'it)]

      [(_ (_ prefix:id)) (syntax-error stx
                                       (message:expect-found message:a-library-path-after-the-prefix
                                                             message:none))]
      ; Correct syntax
      [(_ (_ prefix:id m-id:id)) (check-module-id-form stx #'m-id)
                                 (syntax/loc stx (require (prefix-in prefix m-id)))]
      
      [(_ (_ prefix:id non-id))
       (syntax-error stx (message:expect-found message:a-library-path-after-the-prefix #'non-id)
                     #'non-id)]
      ; Too many terms
      [(_ (_ _:id _:id extra ...))
       (syntax-error stx (message:expect-found
                          message:only-prefix-and-path
                          (message:extra-parts (length (stx->list #'(extra ...))))))]

      [(_ (_ non-id . _)) (syntax-error stx (message:expect-found message:a-name-as-a-prefix #'non-id)
                                        #'non-id)])))

(begin-for-syntax
  (lazy-require [(submod "." error) (⌢:require check-module-form parse-only-in parse-prefix-in)]))

(define-syntax compthink:require
  (parser/⊥
   #:literals (compthink:only-in compthink:prefix-in)
   [(_ (~or* module:str module:id)) #:when (top?) (check-module-form this-syntax #'module) ; ⌢?
                                    #'(require module)]
   ; maybe forgot brackets
   [(~or* (_ compthink:only-in   . _) (_ (compthink:only-in   . _)))
    #:when (top?)(parse-only-in   this-syntax)]
   [(~or* (_ compthink:prefix-in . _) (_ (compthink:prefix-in . _)))
    #:when (top?) (parse-prefix-in this-syntax)]
   [_ (⌢:require this-syntax)]))

; ‘set!-transformer’ was overkill, although it singled out the location as name in (name . _),
;  which we might do later.
(define-syntaxes (compthink:only-in compthink:prefix-in)
  (let ([parser (syntax-parser [_ (syntax-error this-syntax "out of context"
                                                #;message:sub-form-out-of-context)])])
    (values parser parser)))
