#lang racket/base

(module forms racket/base (provide ~s
                                   top:checked-identifier
                                   define:interaction:redefine-check
                                   or/and:checked-boolean if:checked-boolean
                                   true?
                                   for-loop
                                   time
                                   wrap-λ)

  (provide (prefix-out base: (combine-out
                              #%top-interaction
                              #%top
                              begin
                              (rename-out [#%plain-app #%app] [#%plain-lambda λ])
                              quote 
                              and or define cond else if
                              let)))
  
  (require (only-in racket/format ~s))
  
  (require racket/lazy-require)
  (lazy-require [(submod "../../english.rkt" syntax) (define:redefine:user)])
  (define (define:interaction:redefine-check definition id)
    (when (or (identifier-binding id)
              ;; identifier-binding returns #f for variable bound at the top-level, check explicitly:
              (namespace-variable-value (syntax-e id) #true (λ () #false)))
      (syntax-error definition (define:redefine:user id) id)))  
  (define ⌢:top:checked-identifier:unbound "this variable is undefined")
  (define (top:checked-identifier id)
    (namespace-variable-value
     (syntax-e id) #true (λ () (syntax-error id ⌢:top:checked-identifier:unbound))))
  (define (syntax-error stx msg [detail #false]) (raise-syntax-error #false msg stx detail))
  
  ; Checks that conditions are boolean.
  (require racket/lazy-require)
  (lazy-require [(submod "../../english.rkt" error)
                 (expect-received for-the-nth-argument context-description indefinite ~a␣)]
                [racket/format (~e ~s)])
  (define (or/and:checked-boolean name value placement)
    (if (boolean? value) value (boolean-err name value (for-the-nth-argument (add1 placement)))))
  (define (if:checked-boolean name value)
    (if (boolean? value) value (boolean-err name value "from the condition")))
  (define (boolean-err name value from) (err name (expect-received (~a␣ (indefinite "boolean") from)
                                                                   (~e value))))
  (define (err function/name message)
    (raise (make-exn:fail:contract (context-description (the-name function/name) message)
                                   (current-continuation-marks))))
  (define (the-name f) (or (object-name f) f))

  (define anonymous-name '|an anonyomus function|)

  (struct anon (f code)
    #:property prop:procedure   (struct-field-index f)
    #:property prop:object-name (λ (self) (syntax→code ((anon-code self)))))

  (require (only-in "../../shared/code.rkt" syntax→code))
  (define (wrap-λ a-λ closed-code) (anon (procedure-rename a-λ anonymous-name) closed-code))

  (define (true? v) (eq? v #true))

  (define message:ms "ms")

  (define for:accumulators:arity-error
    (string-append "expected "
                   "a list with exactly as many elements as the number of accumulators"))
  (define for:accumulators:value-error
    (string-append "expected "
                   "a list of values to match the list of accumulators"
                   ", but found "
                   "something else"))
  (define for:sequence:value-error ; Wouldn't happen for the recursive iterations.
    (string-append "expected "
                   "a list or text to iterate through"
                   ", but found "
                   "something else"))
  (define (for-loop name s as λ-body single?)
    (unless (or (list? s) (string? s)) (err name for:sequence:value-error))
    (define result (for/fold ([as as]) ([e s])
                     (cond [(not (list? as)) (err name for:accumulators:value-error)]
                           [(not (= (procedure-arity λ-body) (add1 (length as))))
                            (err name for:accumulators:arity-error)]
                           [else (apply λ-body (if (char? e) (string e) e) as)])))
    (if single? (car result) result))

  (define (time thunk expression) ; Run-time support, macro just thunks and quotes.
    (define-values (vs cpu user gc) (time-apply thunk '()))
    (printf "~s ~a : ~s\n" user message:ms expression)
    (apply values vs)))

(module non-image racket/base

  (provide identity
           nullary? unary? binary? ternary? quaternary? arity-range
           differ?
           %
           non-negative?
           element? remove sub-list (rename-out [append′ append])
           text->list sub-text?
           letters? whitespace? lower-case? upper-case? digits?
           character? character character->unicode unicode->character
           repeats cross parallel
           sort)

  (require (only-in racket/unsafe/ops unsafe-cons-list)
           (only-in racket/list take drop)
           (prefix-in base: (only-in racket/base sort)))

  ; HoFs producing a function
  ; —————————————————————————
  (provide compose >∘ •)
  (define (compose . fs) (literal-call (apply compose1          fs)  '∘  fs))
  (define (>∘      . fs) (literal-call (apply compose1 (reverse fs)) '>∘ fs))
  (define (• f    . .as) (literal-call (λ (a) (apply f a .as))       '• (list* f .as)))
  (struct literal-call (ϕ name arguments)    
    #:property prop:procedure (struct-field-index ϕ)
    #:property prop:object-name
    (λ (self) (list* (literal-call-name self) (map (λ (v) (or (object-name v) v))
                                                   (literal-call-arguments self)))))

  ; ———
  (provide >>)
  (define (>> v f) (f v))

  (define (repeats f a n)
    (if (zero? n) '() (let repeats ([a a] [n (sub1 n)] [ℓ (list a)])
                        (cond [(zero? n) (reverse ℓ)]
                              [else (define fa (f a))
                                    (repeats fa (sub1 n) (unsafe-cons-list fa ℓ))]))))

  (define (sort comparator ℓ) (base:sort ℓ comparator))
  
  (require (only-in racket/list cartesian-product))
  (define (cross    hof f l.1 l.2) (hof (λ (l.12) (apply f l.12)) (cartesian-product l.1 l.2)))
  (define (parallel hof f l.1 l.2) (hof (λ (l.12) (apply f l.12)) (map list          l.1 l.2)))

  (define (identity v) v)
  (define (nullary?     f) (procedure-arity-includes? f 0))
  (define (unary?       f) (procedure-arity-includes? f 1))
  (define (binary?      f) (procedure-arity-includes? f 2))
  (define (ternary?     f) (procedure-arity-includes? f 3))
  (define (quaternary?  f) (procedure-arity-includes? f 4))
  (define (arity-range  f) 'ToDo #;procedure-arity)

  (define (differ? v₀ v₁) (not (equal? v₀ v₁)))

  (define (% x) (* x 1/100))

  (define (non-negative? x) (0 . <= . x))
  
  (define (element? e ℓ) (and (member e ℓ) #true))
  (define (remove   e ℓ) (remove* (list e) ℓ))
  (define (sub-list ℓ i j) (take (drop ℓ i) (- j i)))
  (define (append′ ℓ e) (append ℓ (list e)))
  
  ; ISL uses beginner: ternary, checks each argument, then calls range.
  ; 2htdp/abstraction has in-range, which is unary or ternary.
  ; racket/list.rkt
  #;(define range-proc
      (let () ; make sure range has the right runtime name
        (define range (case-lambda
                        [(end)            (for/list ([i (in-range end)])            i)]
                        [(start end)      (for/list ([i (in-range start end)])      i)]
                        [(start end step) (for/list ([i (in-range start end step)]) i)]))
        range))  

  (define (text->list t) (map string (string->list t))) ; ISL
  (define (sub-text? t₀ t) (regexp-match? (regexp-quote t₀) t)) ; ISL

  (define (letters? t)    (andmap char-alphabetic? (string->list t)))
  (define (whitespace? t) (andmap char-whitespace? (string->list t)))
  (define (lower-case? t) (andmap char-lower-case? (string->list t)))
  (define (upper-case? t) (andmap char-upper-case? (string->list t)))
  (define (digits? t)     (andmap char-numeric?    (string->list t)))
  
  (define (character? v) (and (string? v) (= (string-length v) 1)))
  (define (character t i) (string (string-ref t i)))
  (define (character->unicode c) (char->integer (string-ref c 0)))
  (define (unicode->character n) (string (integer->char n))))


(module image racket/base (provide image? freeze bitmap/url
                                   (except-out (all-defined-out)
                                               with-element #;list-set
                                               % byte← normalize))
  (require (only-in mrlib/image-core image?)
           (only-in 2htdp/private/image-more freeze)
           (prefix-in 2htdp: (only-in 2htdp/private/image-more
                                      rotate crop scale scale/xy
                                      image-width image-height
                                      above beside overlay above/align beside/align overlay/align
                                      circle square triangle star rectangle ellipse
                                      text))
           (prefix-in 2htdp:
                      (only-in mrlib/image-core (make-color color)
                               color-red color-blue color-green color-alpha))
           "../../snack/boolean.rkt"
           "../../snack/match.rkt"
           (only-in racket/draw the-color-database)
           (only-in racket/class send) 
           (only-in racket/math exact-round)
           (only-in racket/list take drop first second third fourth))

  (require (only-in net/url call/input-url string->url get-pure-port)
           (only-in racket/draw bitmap%)
           (prefix-in image-more: (only-in 2htdp/private/image-more rotate)))
  (define (bitmap/url string)
    (image-more:rotate ; dummy rotate coerces to a 2htdp/image image
     0 (call/input-url (string->url string)
                       (λ (url [header '()]) (get-pure-port url header #:redirections 5))
                       (λ (port) (make-object bitmap% port 'unknown/alpha #false #true)))))

  (define (byte← percentage) (exact-round (* 2.55 percentage)))
  
  (define (normalize color) (if (list? color) (apply 2htdp:color (map byte← color)) color))
  
  (define 1/ϕ (/ ((sqrt 5) . - . 1) 2))
  (define θ -27)
  (require (only-in racket/math pi))
  (define θ-scale (/ (sqrt 2) (+ (sin (/ pi 5)) (cos (/ pi 5)))))
  (define r (* 1/ϕ θ-scale))

  (define (circle   size [fill "outline"] [color "black"])
    (2htdp:circle   (/ size 2) fill (normalize color)))
  (define (square   size [fill "outline"] [color "black"])
    (2htdp:square   size fill (normalize color)))
  (define (triangle size [fill "outline"] [color "black"])
    (2htdp:triangle size fill (normalize color)))
  (define (star     size [fill "outline"] [color "black"])
    (2htdp:rotate θ (2htdp:star (* size r) fill (normalize color))))

  (define (rectangle width height [fill "outline"] [color "black"])
    (2htdp:rectangle width height fill (normalize color)))
  (define (ellipse   width height [fill "outline"] [color "black"])
    (2htdp:ellipse   width height fill (normalize color)))

  (define (text->image string [font-size 12] [color "black"])
    (2htdp:text string (exact-round font-size) (normalize color)))
           
  (define (color? v)
    (match v
      [(list r g b)   (and (<= 0 r 100) (<= 0 g 100) (<= 0 b 100))]
      [(list r g b α) (and (<= 0 r 100) (<= 0 g 100) (<= 0 b 100) (<= 0 α 100))]
      ["transparent" #true]
      [(? string?) (truth? (send the-color-database find-color v))]
      [_ #false]))

  (define (colors? v) (and (list? v) (andmap color? v)))

  ; Reimplement 2htdp's image->color-list.
  ; Main benefit is avoiding intermediate list of color structs.
  #;(define (percentage← byte) (* 1/10 (exact-round (* byte 1000/255))))
  #;(define (image->colors an-image)
      (map (λ (a-color) (list (% (2htdp:color-red   a-color))
                              (% (2htdp:color-green a-color))
                              (% (2htdp:color-blue  a-color))
                              (% (2htdp:color-alpha a-color))))
           (2htdp:image->color-list an-image)))
  ;
  ; Main speed-up is caching percentage, then not making intermediate list of color structs,
  ;  then not mapping % over the components.
  ;
  (require (only-in racket/draw make-bitmap bitmap-dc%)
           (only-in mrlib/image-core render-image bitmap->image)
           (only-in racket/class make-object))
  ;
  (define % ; Cache the 256 possibilities up-front.
    ; See also ‘rationalize’.
    (let ([percentages (build-vector 256 (λ (byte) (* 1/10 (exact-round (* byte 1000/255)))))])
      (λ (b) (vector-ref percentages b))))
  ;
  (define (image->colors image)
    (define w (2htdp:image-width image))
    (define h (2htdp:image-height image))
    (cond [(or (zero? w) (zero? h)) '()]
          [else (define bdc (make-object bitmap-dc% (make-bitmap w h)))
                (send bdc erase)
                (render-image image bdc 0 0)
                (define bytes (make-bytes (* w h 4)))
                (send bdc get-argb-pixels 0 0 w h bytes)
                ; Loop is 90% of the time, unsafe operations net maybe 5% improvement.
                ; Not doing % gets 10% improvement, so it's the inner list building that's
                ;  extra vs 2htdp, changing to vector saves 40%.
                (for/list ([i (in-range 0 (* w h 4) 4)])
                  ; Simple tests outside this context suggest nested consing is no faster.
                  (list (% (bytes-ref bytes (+ i 1)))
                        (% (bytes-ref bytes (+ i 2)))
                        (% (bytes-ref bytes (+ i 3)))
                        (% (bytes-ref bytes i))))]))
  ; 1/3 of the time is the contract checking
  ; Changing byte← to use truncate and/or #i2.55 makes no detectable difference.
  ; Not going through a map to structs speeds it up by a factor of 5.
  ; (make-bitmap width height [alpha? #:backing-scale backing-scale])
  ;  → (is-a?/c bitmap%)
  ;   width : exact-positive-integer?
  ;   height : exact-positive-integer?
  ;   alpha? : any/c = #t
  ;   backing-scale : (>/c 0.0) = 1.0
  (define (colors->image color-list width height)
    (cond [(or (zero? width) (zero? height)) (rectangle width height "solid" "black")]
          [else (define bytes (make-bytes (* width height 4) 0))
                (for ([c (in-list color-list)]
                      [i (in-naturals)])
                  (define j (* i 4))
                  (match c
                    [(list r g b) (bytes-set! bytes j       255)
                                  (bytes-set! bytes (+ j 1) (byte← r))
                                  (bytes-set! bytes (+ j 2) (byte← g))
                                  (bytes-set! bytes (+ j 3) (byte← b))]
                    [(list r g b α) (bytes-set! bytes j       (byte← α))
                                    (bytes-set! bytes (+ j 1) (byte← r))
                                    (bytes-set! bytes (+ j 2) (byte← g))
                                    (bytes-set! bytes (+ j 3) (byte← b))]
                    [_ (define str (if (string? c) c (symbol->string c)))
                       (define clr (or (send the-color-database find-color str)
                                       (send the-color-database find-color "black")))
                       ;; 2htdp implementation says:
                       ;   this should probably (send clr alpha) when that's possible
                       (bytes-set! bytes j 255)
                       (bytes-set! bytes (+ j 1) (send clr red))
                       (bytes-set! bytes (+ j 2) (send clr green))
                       (bytes-set! bytes (+ j 3) (send clr blue))]))
                (define bmp (make-bitmap width height))
                (send bmp set-argb-pixels 0 0 width height bytes)
                (bitmap->image bmp)]))
  #;(define (bitmap->image bm [mask-bm (send bm get-loaded-mask)])
      (define w (send bm get-width)) (define h (send bm get-height))
      (define alpha-bm (cond [(send bm has-alpha-channel?) bm]
                             [else (define new-bm (make-bitmap w h))
                                   (define bdc (make-object bitmap-dc% new-bm))
                                   (send bdc draw-bitmap bm 0 0 'solid
                                         (send the-color-database find-color "black")
                                         mask-bm)
                                   (send bdc set-bitmap #f)
                                   new-bm]))
      (make-image (make-translate (/ w 2) (/ h 2) (make-ibitmap alpha-bm 0 1 1 (make-hash)))
                  (make-bb w h h) #f))
  #;(define (make-image shape bb normalized? [pinhole #f])
      (new image% [shape shape] [bb bb] [normalized? normalized?] [pinhole pinhole]))
  ;; a bitmap is:
  ;;  - (make-ibitmap (and/c (is-a?/c bitmap%) (lambda (x) (send x has-alpha-channel?)))
  ;;                  angle positive-real 
  ;;                  hash[(list boolean[flip] number[x-scale] number[y-scale] number[angle])
  ;;                       -o> (is-a?/c bitmap%)])
  ;;    NOTE: bitmap copying needs to happen in 'write' and 'read' methods
  #;(define-struct/reg-mk ibitmap #:reflect-id bitmap (raw-bitmap angle x-scale y-scale cache)
      #:omit-define-syntaxes #:transparent
      #:property prop:custom-write (λ (x y z) (bitmap-write x y z)))
  
  (define (with-element ℓ i e) (append (take ℓ i) (list e) (drop ℓ (add1 i))))
  (define red     (case-lambda [(color) (first  color)] [(color %) (with-element color 0 %)]))
  (define green   (case-lambda [(color) (second color)] [(color %) (with-element color 1 %)]))
  (define blue    (case-lambda [(color) (third  color)] [(color %) (with-element color 2 %)]))
  (define opacity (case-lambda [(color) (fourth color)] [(color %) (with-element color 3 %)]))
  
  #;(define (colors->image colors width height)
      (2htdp:color-list->bitmap (map normalize colors) width height))

  (define (blank [size₁ 0] [size₂ size₁]) (rectangle size₁ size₂ "solid" "transparent"))
  
  (define (scale   image factor) (if (zero? factor) (blank) (2htdp:scale factor image)))
  (define (shrink  image) (2htdp:scale 0.5 image))
  (define (enlarge image) (2htdp:scale 2   image))
  ;
  (define (scale-width image factor)
    (if (zero? factor) (blank 0 (2htdp:image-height image)) (2htdp:scale/xy factor 1 image)))
  (define (scale-height image factor)
    (if (zero? factor) (blank (2htdp:image-width image) 0)  (2htdp:scale/xy 1 factor image)))
  (define (thinner image) (scale-width  image 0.5))
  (define (wider   image) (scale-width  image 2))
  (define (shorter image) (scale-height image 0.5))
  (define (taller  image) (scale-height image 2))
  
  (define (rotate    image angle) (2htdp:rotate angle image))
  (define (counter   image)       (2htdp:rotate  90   image))
  (define (clockwise image)       (2htdp:rotate -90   image))

  (define empty-image (2htdp:square 0 'solid 'black)) ; square a derived concept?
  (define beside  (case-lambda [() empty-image] [(img) img] [images (apply 2htdp:beside  images)]))
  (define above   (case-lambda [() empty-image] [(img) img] [images (apply 2htdp:above   images)]))
  (define overlay (case-lambda [() empty-image] [(img) img] [images (apply 2htdp:overlay images)]))
  ; ToDo: case-lambda.
  (define (align-above   x   . images) (apply 2htdp:above/align   x
                                              empty-image empty-image images))
  (define (align-beside    y . images) (apply 2htdp:beside/align    y
                                              empty-image empty-image images))
  (define (align-overlay x y . images) (apply 2htdp:overlay/align x y
                                              empty-image empty-image images))
  
  (define (cut-left image pixels)
    (let ([pixels (exact-round pixels)])
      (2htdp:crop pixels 0
                  (- (2htdp:image-width image) pixels)
                  (2htdp:image-height image) image)))
  (define (cut-top image pixels)
    (let ([pixels (exact-round pixels)])
      (2htdp:crop 0 pixels
                  (2htdp:image-width image)
                  (- (2htdp:image-height image) pixels) image)))
  (define (cut-right image pixels)
    (2htdp:crop 0 0
                (- (2htdp:image-width image) (exact-round pixels))
                (2htdp:image-height image) image))
  (define (cut-bottom image pixels)
    (2htdp:crop 0 0
                (2htdp:image-width image)
                (- (2htdp:image-height image) (exact-round pixels)) image))
  ;
  (define (left   image pixels) (let ([pixels (exact-round pixels)])
                                  (2htdp:crop 0 0 pixels (2htdp:image-height image) image)))
  (define (top    image pixels) (let ([pixels (exact-round pixels)])
                                  (2htdp:crop 0 0 (2htdp:image-width image) pixels  image)))
  (define (right  image pixels) (let ([pixels (exact-round pixels)])
                                  (2htdp:crop (- (2htdp:image-width image) pixels) 0
                                              pixels (2htdp:image-height image)
                                              image)))
  (define (bottom image pixels) (let ([pixels (exact-round pixels)])
                                  (2htdp:crop 0 (- (2htdp:image-height image) pixels)
                                              (2htdp:image-width image) pixels
                                              image))))
