#lang racket/base (require (submod "meta-language.rkt" non-image))

(define-for-syntax default #'racket/base)

(define-syntax-parser create 
  [(_ :contracting binding)
   #:declare binding (alias/rename/from #'f #'here custom)
   (define/provide this-syntax (attribute binding.library) default #'binding.data)])

(module* nullary #false
  (require syntax/parse)
  (provide (prefix-out non-image: nullary-id?))
  (define-literal-set nullary (+ * text-join list join))
  (define nullary-id? (literal-set->predicate nullary)))

; Build some checks from contracts.rkt contracts via error.rkt combinators
; ————————————————————————————————————————————————————————————————————————
(define check:range (arity-arguments-contract-checker 1 number/C number/C non-zero/C))
(define check:-     (arity-arguments-contract-checker 1 number/C number/C))
(define-values (check-numbers check-texts check-lists check-unaries check-functions)
  (values (arguments-homogeneous-contract-checker number/C)
          (arguments-homogeneous-contract-checker text/C)
          (arguments-homogeneous-contract-checker list/C)
          (arguments-homogeneous-contract-checker unary/C)
          (arguments-homogeneous-contract-checker function/C)))
(define check:function-unaries (checker-fixed-homogeneous `(,function/C) unary/C))
(define-values (check-list-index check-text-index)
  (values (binary-checker list-index/C) (binary-checker text-index/C)))
(define-values (check-list-range check-text-range)
  (values (dependent-argument-checker `(2 ,list-lower/C) `(3 ,list-upper/C) `(3 ,list-order/C))
          (dependent-argument-checker `(2 ,text-lower/C) `(3 ,text-upper/C) `(3 ,text-order/C))))
(define check:partial (checker-fixed-homogeneous `(,function/C ,any/C) any/C))

(map-macro create

           ; Type Predicate
           ; ——————————————
           
           ; Early Lectures
           [(number?   any/C) #:renames real?]
           [(function? any/C) #:renames procedure?]
           [(boolean?  any/C)]
           [(text?     any/C) #:renames string?]
           [(list?     any/C)]           

           ; List
           ; ————
           
           ; Early Lectures
           [list]
           [(length list/C)]
           [(map unary/C list/C)]
           [(apply function/C list/C) [#:aka combine]]
           [(range x₁ . range-spec) check:range #:from racket/list]
           [(list-of natural/C any/C) #:renames make-list #:from racket/list]
           
           [(first non-empty/C) #:renames unsafe-car #:from racket/unsafe/ops]
           [(rest  non-empty/C) #:renames unsafe-cdr #:from racket/unsafe/ops]
           [(second two/C)   #:renames cadr]
           [(third  three/C) #:renames caddr]
           [(fourth four/C ) #:renames cadddr]
           [(empty? list/C)  #:renames null?]
           [(join . ℓs) check-lists #:renames append]
           ; ★
           [(adjoin any/C list/C) [#:aka prepend] #:renames unsafe-cons-list #:from racket/unsafe/ops]
           [(append list/C any/C) #:custom]
           [(select unary-predicate/C list/C) [#:aka sift only] #:renames filter] ; ★
           [(sort  binary-predicate/C list/C) #:custom]
           ;
           [(remove   any/C list/C) #:custom]
           [(element? any/C list/C) [#:aka ∈] #:custom]
           [(element  list/C natural/C)           check-list-index #:renames list-ref]
           [(sub-list list/C natural/C natural/C) check-list-range #:custom]
           [(reverse list/C)]
           ;
           [(repeats unary/C any/C natural/C) #:custom]
           
           ; Text
           ; ————
           
           ; Early Lectures
           [(text-length text/C) #:renames unsafe-string-length #:from racket/unsafe/ops]
           [(text-join . ts) check-texts #:renames string-append]

           [(character text/C natural/C)           check-text-index #:custom]
           [(sub-text  text/C natural/C natural/C) check-text-range #:renames substring]
           [(sub-text? text/C text/C) #:custom]
           [(number->text number/C) [#:aka number→text] #:renames number->string]
           [(text->number text/C)   [#:aka text→number] #:renames string->number]
           [(text->list   text/C) [#:aka text→list] #:custom]
           [(lower-case?  text/C) #:custom]
           [(upper-case?  text/C) #:custom]
           [(letters?     text/C) #:custom]
           [(digits?      text/C) #:custom]
           [(whitespace?  text/C) #:custom]
           [(lower-case   text/C) #:renames string-downcase]
           [(upper-case   text/C) #:renames string-upcase]
           [(character->unicode character/C) [#:aka character→unicode] #:custom]
           [(unicode->character natural/C)   [#:aka unicode→character] #:custom]
           [(alphabet-order? t₀ t₁ . ts) check-texts #:renames string<=?]
           [(character? any/C) #:custom]

           ; Number
           ; ——————

           ; Early Lectures
           [pi #:from racket/math] [π #:renames pi #:from racket/math]
           [(+ . xs) check-numbers]
           [(* . xs) check-numbers]
           [(- x₀ . xs) check:-]
           [(/ number/C non-zero/C)]
           [(inc number/C) #:renames add1]
           [(dec number/C) #:renames sub1]
           
           [(sin number/C) [#:aka sine]]
           [(cos number/C) [#:aka cosine]]

           [(% number/C) #:custom]

           [(=  x₀ x₁ . xs) check-numbers]
           [(<  x₀ x₁ . xs) check-numbers]
           [(>  x₀ x₁ . xs) check-numbers]
           [(<= x₀ x₁ . xs) check-numbers [#:aka ≤]]
           [(>= x₀ x₁ . xs) check-numbers [#:aka ≥]]
           [(minimum x . xs) check-numbers #:renames min]
           [(maximum x . xs) check-numbers #:renames max]
           
           [(integer? any/C) #:renames exact-integer?]
           [(natural? any/C) #:renames exact-nonnegative-integer?]
           [(negative?     number/C)]
           [(zero?         number/C)]
           [(positive?     number/C)]
           [(non-negative? number/C) #:custom]
           [(squared number/C)           #:renames sqr #:from racket/math]
           [(power positive/C number/C)  #:renames expt]
           [(square-root non-negative/C) [#:aka √] #:renames sqrt]
           [(remainder natural/C positive-natural/C)]
           [(quotient  natural/C positive-natural/C)]
           [(even? integer/C)]
           [(odd?  integer/C)]
           [(round   number/C) #:renames exact-round   #:from racket/math]
           [(floor   number/C) #:renames exact-floor   #:from racket/math]
           [(ceiling number/C) #:renames exact-ceiling #:from racket/math]

           [(random small-enough-for-random/C) #:from '#%kernel]
           
           ; Function
           ; ————————
           
           #;[(nullary?    function/C) #:custom]
           [(unary?      function/C) #:custom]
           [(binary?     function/C) #:custom]
           #;[(ternary?    function/C) #:custom]
           #;[(quaternary? function/C) #:custom]
           #;[(arity-range function/C) #:custom]

           ; Boolean
           ; ———————
           
           [(not boolean/C)]
           
           ; Other
           ; —————
           
           [(same?   any/C any/C) #:renames equal?]
           [(differ? any/C any/C) #:custom]
   
           [(identity any/C) #:custom]

           [(url->browser text/C) [#:aka url→browser] #:renames send-url #:from net/sendurl]

           ; ———
           
           [(>> any/C unary/C) #:custom]
           
           ; HoFs producing a function
           ; —————————————————————————           
           [(compose f₀ f₁ . fs) #;check-functions check-unaries [#:aka ∘] #:custom]
           [(>∘      f₀ f₁ . fs) check:function-unaries #:custom]           
           [(• f b . .bs) check:partial #:custom]

           ; ———
           [(cross    function/C binary/C list/C list/C) #:custom]
           [(parallel function/C binary/C list/C list/C) #:custom])
