#lang racket/base (require racket/require)

(require "../../snack/boolean.rkt"
         (submod "../../snack/contract.rkt" core)
         "../../snack/preferences.rkt"
         (submod "../../snack/list.rkt" core))

(require racket/unit
         (only-in racket/string string-prefix?)
         (subtract-in framework "../../snack/preferences.rkt")
         racket/local)

(require (submod "../../snack/class.rkt" implement)
         "../../snack/gui.rkt")

(require "gui.rkt"
         "range-rectangle.rkt"
         (only-in "struct.rkt" define/extract-struct)
         (only-in "values.rkt" values₃:0)
         (only-in "class.rkt" define/extend define/augment* send/values)
         racket/gui/base)

(require syntax/parse/define (for-syntax racket/base))
(define-simple-macro (list/cond* id [element:expr test:expr ...+] ...+)
  #:declare id id #:role "list accumulation variable"
  (do-list/cond* (list (λ (id) test ...) ...) (list (λ (id) element) ...)))
(define (do-list/cond* λ-tests λ-elements)
  (for/lists (id) ([λ-test λ-tests] [λ-element λ-elements] #:when (λ-test id)) (λ-element id)))

(require drracket/tool) (provide tool@)

(define-values (shape-preference detail-preference enabled-preference)
  (values 'framework:boxed-highlight-scheme
          'framework:boxed-highlight-details
          'framework:boxed-highlight-enabled?))
(define-values (shape/! detail/! enabled/!)
  (values (make-parameter null)
          (make-parameter null)
          (make-parameter null)))

(define tool@
  (unit (import drracket:tool^)
        (export drracket:tool-exports^)
        
        (preferences:set-default shape-preference 'box (or/c 'line 'box-truncated 'box))
        (preferences:set-default detail-preference '(outline inset-horizontal inset-vertical)
                                 (listof (or/c 'outline 'inset-horizontal 'inset-vertical)))
        (preferences:set-default enabled-preference #t boolean?)
        (shape/!   (preferences:get shape-preference))
        (detail/!  (preferences:get detail-preference))
        (enabled/! (preferences:get enabled-preference))

        (drracket:get/extend:extend-unit-frame
         (mixin (drracket:unit:frame<%>) () (super-new)
           (inherit set-show-menu-sort-key get-current-tab)
           (define/extend (add-show-menu-items show-menu) #:super-last
             (with-parent show-menu
               (define boxed-highlight-menu-item (new menu% [label "Boxed Highlight"]))
               (with-parent boxed-highlight-menu-item
                 (define sub-menus
                   (for/list ([label '("Standard DrRacket" "Truncated Box" "Rounded Box")]
                              [mode '(line box-truncated box)])
                     (new menu:can-restore-checkable-menu-item%
                          [label label] [checked (eq? mode (shape/!))]
                          [callback (λ (self _)
                                      (for-each (λ (sub) (send sub check (eq? sub self))) sub-menus)
                                      (preferences:set shape-preference mode))])))
                 (define (detail-item detail label)
                   (new menu:can-restore-checkable-menu-item%
                        [label label] [checked (memq detail (detail/!))]
                        [callback (λ (self _)
                                    (define details (detail/!))
                                    (preferences:set detail-preference
                                                     ((if (memq detail details) remove* append)
                                                      (list detail) details)))]))
                 (new separator-menu-item%)
                 (detail-item 'inset-horizontal "Horizontally Inset Boxes")
                 (detail-item 'inset-vertical "Vertically Inset Boxes")
                 (detail-item 'outline "Outline [Rounded Box Only]"))
               (set-show-menu-sort-key boxed-highlight-menu-item -1)))

           ; Disable for scribble files.
           ;
           (define/public (boxed-highlight-enabled?)
             (let* ([definitions-text (send (get-current-tab) get-defs)]
                    [settings (send definitions-text get-next-settings)]
                    [name (extract-language-name settings definitions-text)])
               (not (string-prefix? (string-downcase name) "scribble"))))
           ;
           ; Taken from drracket/drracket/private/rep.rkt.
           (define (extract-language-name language-settings defs-text)
             (cond
               [(is-a? (drracket:language-configuration:language-settings-language language-settings)
                       drracket:module-language:module-language<%>)
                (send (drracket:language-configuration:language-settings-language language-settings)
                      get-users-language-name defs-text
                      (drracket:language-configuration:language-settings-settings language-settings))]
               [else
                (send (drracket:language-configuration:language-settings-language language-settings)
                      get-language-name)]))))
        
        (define (phase1) (void))
        
        (define (phase2)
          (drracket:get/extend:extend-definitions-text
           (compose boxed-highlight-mixin editor<%>-mixin))
          (drracket:get/extend:extend-interactions-text
           (compose boxed-highlight-mixin editor<%>-mixin)))

        (define (toggle-enabled ed evt) (preferences:set enabled-preference (not (enabled/!))))

        (send (keymap:get-global) add-function "toggle boxed highlight" toggle-enabled)
        #;(send (keymap:get-global) map-function "c:s:&" "toggle boxed highlight")))

(define check-highlight-interface<%>
  (interface (text:basic<%>)
    [highlight-range
     (->i ([this any/c]
           [start natural-number/c] [end (start) (and/c natural-number/c (>=/c start))]
           [color (or/c string? (is-a?/c color%))])
          ([caret-space? boolean?]
           [priority (or/c 'low 'high)]
           [style (or/c 'rectangle 'dot 'ellipse 'hollow-ellipse)]
           #:adjust-on-insert/delete? [adjust-on-insert/delete? boolean?]
           #:key [key any/c])
          [_ any/c])]
    rehighlight-ranges))

(define boxed-highlight-mixin
  (mixin (text:basic<%> editor<%>-mixin-interface) (text:basic<%> check-highlight-interface<%>)
    
    (inherit get-position-values get-top-level-window get-highlighted-ranges)

    (define/public (rehighlight-ranges)
      (define ranges (get-highlighted-ranges))
      (unhighlight-ranges (λ _ #t))
      (for-each
       (λ (r)
         (highlight-range (text:range-start r) (text:range-end r) (text:range-color r)
                          (text:range-caret-space? r) 'low 'rectangle))
       ranges))
    
    (define/override (highlight-range start end in-color
                                      [caret-space? #f] [priority 'low] [style 'rectangle]
                                      #:adjust-on-insert/delete? [adjust-on-insert/delete? #f]
                                      #:key [key #f])

      (define tlw (get-top-level-window))
      (cond [(and (enabled/!)
                  tlw (send tlw boxed-highlight-enabled?)
                  (eq? style 'rectangle) (>= (- end start) 2)
                  (call-with-values (λ () (get-position-values)) =)
                  (neither (eq? (shape/!) 'line)
                           key adjust-on-insert/delete?))
             
             (define color (♯as-color% in-color))
             (unless color
               (error
                'highlight-range
                "expected a color or a string in the-color-database for the third argument, got ~e"
                in-color))
             
             (define l (range start end caret-space? color #f))
             (apply (if (eq? priority 'high) enqueue! enqueue-front!) (list ranges-deq l))
             (set-range-rectangles! l (compute-rectangles l))
             (invalidate-rectangles (range-rectangles l))
             (λ () (unhighlight-range start end color caret-space? 'rectangle))]
            
            [else (super highlight-range start end in-color caret-space? priority style
                         #:adjust-on-insert/delete? adjust-on-insert/delete?
                         #:key key)]))
    
    (local-require data/queue)
    
    (define/augment* (on-reflow) #:inner-last
      (recompute-range-rectangles))
    (define/augment* (after-load-file success?) #:inner-first
      (when success? (set! ranges-deq (make-queue))))
    
    (define-values (ranges-deq pending-ranges recompute-callback-running?)
      (values (make-queue) '() #f))
    (define (recompute-range-rectangles)
      (set! pending-ranges (queue->list ranges-deq))
      (unless recompute-callback-running?
        (set! recompute-callback-running? #t)
        (queue-callback (λ () (run-recompute-range-rectangles)) #f)))
    
    (inherit get-admin)
    (define (run-recompute-range-rectangles)
      (when (get-admin)
        (define done-time (+ (current-inexact-milliseconds) 20))
        (define (done-time?) (>= (current-inexact-milliseconds) done-time))
        (let loop ([rectangular₀/false #f] [did-something? #f]) 
          (cond [(and did-something? (done-time?))
                 (final-invalidate rectangular₀/false)
                 (queue-callback (λ () (run-recompute-range-rectangles)) #f)]
                [(empty? pending-ranges)
                 (final-invalidate rectangular₀/false)
                 (set! recompute-callback-running? #f)]
                [else
                 (define a-range (first pending-ranges))
                 (set! pending-ranges (rest pending-ranges))
                 (define old-rectangles (range-rectangles a-range))
                 (if old-rectangles
                     (let ([new-rectangles (compute-rectangles a-range)])
                       (loop (cond [(equal? new-rectangles old-rectangles)
                                    rectangular₀/false]
                                   [else
                                    (set-range-rectangles! a-range new-rectangles)
                                    (join-rectangles*
                                     rectangular₀/false
                                     (append old-rectangles new-rectangles))])
                             #t))
                     (loop rectangular₀/false #f))]))))
    
    (define (invalidate-rectangles rectangles) (final-invalidate (join-rectangles* #f rectangles)))
    
    (preferences:add-callback shape-preference (λ (_ v) (shape/! v)
                                                 (rehighlight-ranges)))
    (preferences:add-callback detail-preference (λ (_ v) (detail/! v)
                                                  (recompute-range-rectangles)
                                                  (rehighlight-ranges)))
    (preferences:add-callback enabled-preference (λ (_ v) (enabled/! v)
                                                   (rehighlight-ranges)))
    (define (truncated?) (eq? (shape/!) 'box-truncated))
    (define (inset-horizontal?) (member 'inset-horizontal (detail/!)))
    (define (inset-vertical?) (member 'inset-vertical (detail/!)))
    (define (outline?) (and (not (truncated?)) (member 'outline (detail/!))))
    
    (define bad-indent-color (make-object color% "Salmon"))
    (inherit #;editor:get-current-preferred-font-size
             position-locations-values
             position-locations position-location
             position-line position-paragraph
             last-position
             line-start-position line-length
             paragraph-start-position paragraph-end-position
             position&line->relative-position line&relative-position->position
             get-text visual-side/false)
    
    (define (compute-rectangles a-range)
      (define/extract-struct (start end) (range a-range))
      (let*-values ([(start-eol? end-eol?) (values #f #t)]
                    [(start-paragraph end-paragraph) (values (position-paragraph start start-eol?)
                                                             (position-paragraph end end-eol?))]
                    [(start-line end-line) (values (position-line start start-eol?)
                                                   (position-line end end-eol?))]
                    [(start-x start-y-top _) (position-locations-values start start-eol?)]
                    [(end-x end-y-top end-y-bottom) (position-locations-values end end-eol?)]
                    [(start-x) (local [(define/extract-struct (caret-space?) (range a-range))]
                                 (if caret-space? (add1 start-x) start-x))]
                    [(end-x) (max start-x end-x)])
        (define (build left top right bottom right-shrink? top-shrink? bottom-shrink? color)
          (build-rectangle left top right bottom
                           (let* ([leading (get-text start (add1 start))])
                             (list/cond* style
                                         ['rounded (and (not (truncated?)) (equal? "(" leading))]
                                         ['left-shrink (member leading '("(" "[" "{"))]
                                         ['right-shrink (and right-shrink?
                                                             (member 'left-shrink style))]
                                         ['bottom-shrink bottom-shrink?]
                                         ['top-shrink top-shrink?]))
                           color))
        (define/extract-struct (color) (range a-range))
        (if (= start-line end-line)
            (list (build start-x start-y-top end-x end-y-bottom #t #t #t color))
            (let ([bad-indent?
                   (let ([min-indent (add1 (position&line->relative-position start start-line))])
                     (for/or ([paragraph (in-range (add1 start-paragraph) (add1 end-paragraph))])
                       (define paragraph-start-position₀ (paragraph-start-position paragraph))
                       (regexp-match? "[^ ]" (get-text paragraph-start-position₀
                                                       (min
                                                        (+ min-indent paragraph-start-position₀)
                                                        (paragraph-end-position paragraph))))))]
                  [right-x
                   (for*/fold ([right-x -inf.0])
                              ([line (in-range start-line end-line)]
                               [right-position (in-value (visual-side/false line #f))]
                               #:when right-position)
                     (max right-x (values₃:0 (position-locations-values right-position end-eol?))))])
              (let ([color (if bad-indent? bad-indent-color color)])
                (if (truncated?)
                    (list (build start-x start-y-top (max right-x end-x) end-y-top
                                 (< right-x end-x) #t #f color)
                          (build start-x end-y-top end-x end-y-bottom
                                 #t #f #t color))
                    (list (build start-x start-y-top (max right-x end-x) end-y-bottom
                                 (< right-x end-x) #t #t color))))))))
    
    (inherit unhighlight-range unhighlight-ranges/key)
    (define/override (unhighlight-ranges pred [just-one? #f])
      (define rectangular₀/false #f)
      (define found-one? #f)
      (queue-filter!
       ranges-deq
       (λ (a-range)
         (or (and just-one? found-one?)
             (not (local [(define/extract-struct (start end color caret-space?) (range a-range))]
                    (pred start end color caret-space? 'rectangle #f #f)))
             (begin (set! found-one? #t)
                    (set! rectangular₀/false (join-rectangles* rectangular₀/false
                                                               (range-rectangles a-range)))
                    (set-range-rectangles! a-range #f)
                    #f))))
      (final-invalidate rectangular₀/false)
      (unless (and just-one? found-one?)
        (super unhighlight-ranges pred just-one?)))
    
    (inherit invalidate-bitmap-cache)
    (define (final-invalidate rectangular₀/false)
      (when (and rectangular₀/false (not (trivial? rectangular₀/false)))
        (send/values this invalidate-bitmap-cache (x-y-width-height rectangular₀/false))))
    
    (define-values (border-width left-shrink right-shrink top-shrink) (values .02 -.38 -.3 -.1))
    (define/extend
      (on-paint before? dc paint-left paint-top paint-right paint-bottom dx dy draw-caret)
      #:super-last
      (when before?
        (local-context
         dc ([get-pen set-pen] [get-brush set-brush] [get-smoothing set-smoothing])
         (define (∝ v) (* v (editor:get-current-preferred-font-size)))
         (for* ([a-range (in-queue ranges-deq)] [rectangle₀ (range-rectangles a-range)])
           
           ; ?
           (define/extract-struct (left top right bottom) (rectangle rectangle₀))
           (define rectangular₀ (rectangular left top #:x right #:y bottom))
           (when (intersect? rectangular₀
                             (rectangular paint-left paint-top #:x paint-right #:y paint-bottom))
             (define/extract-struct (style color) (rectangle rectangle₀))
             
             (define border (let-values ([(smoothing border outline)
                                          (if (outline?)
                                              (values 'smoothed (∝ border-width) 'hilite)
                                              (values 'aligned 0. 'transparent))])
                              (send* dc
                                (set-smoothing smoothing)
                                (set-pen color border outline)
                                (set-brush color 'solid))
                              border))
             ; transposed sequences
             (let*-values
                 ([(left top width height)
                   (x-y-width-height
                    (for/fold ([rectangular₀ (shift rectangular₀ dx dy)])
                              ([test (list (and (inset-horizontal?) (member 'left-shrink style))
                                           (and (inset-horizontal?) (member 'right-shrink style))
                                           (and (inset-vertical?) (member 'top-shrink style))
                                           (outline?))]
                               [resize-spec `((x₀ ,(∝ left-shrink))
                                              (x₁ ,(∝ right-shrink))
                                              (y₀ ,(∝ top-shrink))
                                              (centered ,(sub1 (- border)) ,(sub1 (- border))))]
                               #:when test)
                      (apply resize rectangular₀ resize-spec)))]
                  [(width height) (values (max 0 width) (max 0 height))])
               (if (member 'rounded style)
                   (send dc draw-rounded-rectangle left top width height (∝ .4))
                   (send dc draw-rectangle left top width height))))))))
    
    (super-new)))