#lang racket/base #| Replacement for racket/class |#

#| Debugging of classes and objects |#

(module inspect racket/base
  (require racket/class)
  (provide print-class-info print-object)
  (define (print-class-info %)
    (define-values (name field-cnt field-name-list field-accessor field-mutator super-class skipped?)
      (class-info %))
    (printf "class: ~a\n" %)
    (printf "• class name : ~v\n" name)
    (printf "• super-class: ~v ~a\n" super-class (if skipped? "(not immediate)" ""))
    (printf "• methods    : ~v\n" (interface->method-names (class->interface %)))
    (printf "• ~a fields  : ~v\n" field-cnt field-name-list))
  (define (print-object o)
    (printf "object: ~a\n" o)
    (printf "• field names   : ~v\n" (field-names o))
    (printf "• representation: ~v\n" (object->vector o))))

#| Two submodules to require to signal intent, and get an idea of what we commonly use:
     ‘use’ and ‘implement’, with the latter also reproviding the former. |#

(module use racket/base (provide (all-from-out racket/class))
  (require (only-in racket/class
                    is-a?/c
                    is-a?
                    new make-object instantiate
                    send send* send+
                    get-field)))

(module implement racket/base (provide (all-from-out (submod ".." use) racket/class))
  
  (require (submod ".." use)
           
           (only-in racket/class
                    
                    is-a?/c implementation?
                    
                    class->interface

                    interface interface*
                    mixin mixin-contract make-mixin-contract
                    class class* this%

                    inherit inherit/super inherit-field
                    super-new super-make-object
                    super

                    this
                    init init-field field

                    define/override define/public
                    define/augment inner)))

(provide send∗
         send♯
         send/with send♯/with
         send/return
         (all-from-out "false.rkt")
         (except-out (all-from-out 'use) send*))

(require 'use (only-in "false.rkt" let*♯ let*/else app♯))

#| Variants of ‘send’.

 ‘send∗’ replacing ‘send*’, allowing fixed method-name and initial non-compound argument expressions.

 ‘send♯’ that short-circuits and propagates any #false from the object or argument expressions.

    If the object or any of the arguments evaluate to #false, short-circuit and produce #false.
    Otherwise, call the method with the already-evaluated object and arguments.

    Often used to chain text methods taking positions, which take a number but return either
     a number or #false.

 ‘send/with’ and ‘send♯/with’ for nested method calls with a fixed object.

   More specifically, nest zero or more unary method calls around an arbitrary-arity method call.

   Often used to chain text methods taking positions as arguments, e.g.

     (send♯/with a-text (get-forward-sexp (get-forward-sexp (get-forward-sexp position))))

    is almost the same as

     (send♯ a-text get-forward-sexp
            (send♯ a-text get-forward-sexp
                  (send♯ a-text get-forward-sexp position)))

    except we evaluate the object expression at most once. |#

(require (submod "syntax.rkt" core)
         (for-syntax (submod "syntax.rkt" syntax-classes)
                     racket/local))

(define-macro (send♯ object:expr method:id argument:expr ...)
  ; No-argument case can come from ‘send♯/with’.
  #:with (id ...) (generate-temporaries #'(argument ...))
  (let*♯ ([object-value object] [id argument] ...) (send object-value method id ...)))

(define-syntax send∗ (local [(syntax-class (non-compound (~not (_ ...))))]
                       (syntax-parser [(_ object:expr method:id argument:non-compound ...
                                          (~multiple (argument-group ...)))
                                       #:with (id ...) (generate-temporaries #'(argument ...))
                                       #'(let ([object-value object] [id argument] ...)
                                           (send* object-value
                                             (method id ... argument-group ...)
                                             ...))]
                                      [(_ . parts) (syntax/loc this-syntax (send* . parts))])))

(define-syntaxes (send/with send♯/with)
  (local [(syntax-class ((nested delegate)
                         (form object:expr (method:id (method′:id argument ...)))
                         #:with propagated
                         #`(let ([object-value object])
                             (#,delegate object-value method
                                         (form object-value (method′ argument ...))))))
          (define (sender delegate)
            (syntax-parser
              [nesting #:declare nesting (nested delegate) #'nesting.propagated]
              [(_ object:expr (method:id argument ...)) #`(#,delegate object method argument ...)]))]
    (values (sender #'send) (sender #'send♯))))

(define-macro (send/return o:expr method:id arg ...)
  (let ([o-value o]) (send o-value method arg ...) o-value))
