#lang racket/base (require (submod "contract.rkt" core))

(provide valid-keybinding?
         ;
         (contract-out [mac-command (string? . → . string?)])
         alt/option default-modifier get-default-shortcut-prefix
         ;
         global user (rename-out [keymap:get-global global-keymap])
         find-keymap-for-function
         add-binding/s make-replaceable-keybinding
         (all-from-out "menu.rkt"))

(require (only-in racket/gui/base keymap% get-default-shortcut-prefix)
         (only-in framework keymap:get-global keymap:get-user keymap:aug-keymap%)
         (only-in "menu.rkt" remove-shortcut set-shortcut)
         (submod "class.rkt" use)
         "iteration.rkt"
         "lambda.rkt"
         "string.rkt"
         (only-in racket/list flatten)
         racket/local)

(module* racket #false (provide keymap:aug-keymap% racket:get-keymap)
  (require (only-in framework keymap:aug-keymap% racket:get-keymap)))

; Is v is a valid keybinding string for keymaps?
(define (valid-keybinding? v)
  (with-handlers ([exn:fail? (λ₁ #false)]) (send (new keymap%) map-function v "") #true))

#| Modifiers |#

; s: — All platforms: Shift
; c: — All platforms: Control
; a: — Mac OS: Option
; m: — Windows: Alt
;         Unix: Meta
;       Mac OS: Command, when map-command-as-meta-key produces #true.
; d: — Mac OS: Command

(define (mac-command key) (text-join "d:" key))
(define (alt/option keybinding) (map (λ• (text-join • keybinding)) '("m:" "a:")))

; Figure out usual platform-specific modifier key, e.g. usually Ctrl for Windows, Command for Mac ....
; get-default-shortcut-prefix: (case (system-type)
;                                [(windows) (list/c 'ctl)] [(macosx) (list/c 'cmd)]
;                                [(unix) (list/c (or/c 'alt 'cmd 'meta 'ctl 'shift 'option))])
(define default-modifier
  (local [(define modifier/s
            (text-join-map (λ-case [(ctl) "c:"] [(cmd) "d:"] [(shift) "s:"] [(option) "a:"]
                                   [(alt meta) "~c:m:"])
                           (get-default-shortcut-prefix)))]
    (λ• (text-join modifier/s •))))

#| ——— |#

(define-values (global user) (values 'global 'user))
;
(define resolve-keymap (λ-case/identity [(global) (keymap:get-global)] [(user) (keymap:get-user)]))
;
; Add a keybinding with a name and handler to:
;   • a specified keymap, or
;   • the keymaps loaded by ‘keymap:add-user-keybindings-file’, or
;   • the global keymap
(define (add-binding/s global/user/keymap keybinding/s name handler)
  (define keymap (resolve-keymap global/user/keymap))
  (send keymap add-function name handler)
  (for ([keybinding (flatten keybinding/s)]) (send keymap map-function keybinding name)))

#;{(send a-keymap:aug-keymap get-chained-keymaps) → (listof (is-a?/c keymap%))}
{define (find-keymap-for-function keymap name)
  (for/first k (cons keymap (send keymap get-chained-keymaps))
    #:when (send k is-function-added? name))}

; For creating a keybinding that can be replaced.
; The keymap classes provide ways to add keybindings, but not to remove nor replacing existing ones.
; To get around this we create a new keymap with only one keybinding and chain it to the keymap.
; The returned function can be called with a replacement keybinding.
(define (make-replaceable-keybinding global/user/keymap keybinding name handler)
  (let ([keymap (resolve-keymap global/user/keymap)]
        [current-keymap (new keymap:aug-keymap%)])
    (add-binding/s current-keymap keybinding name handler)
    (send keymap chain-to-keymap current-keymap #true)
    (λ (new-keybinding)
      (let ([new-keymap (new keymap:aug-keymap%)])
        (add-binding/s new-keymap new-keybinding name handler)
        (send* keymap
          [remove-chained-keymap current-keymap]
          [chain-to-keymap new-keymap #true])
        (set! current-keymap new-keymap)))))
